"""usdb_syncer's GUI"""

import webbrowser
from collections.abc import Callable
from pathlib import Path

from PySide6 import QtGui
from PySide6.QtWidgets import QFileDialog, QLabel, QMainWindow

from usdb_syncer import SongId, db, events, settings, song_routines, usdb_id_file
from usdb_syncer.constants import Usdb
from usdb_syncer.gui import gui_utils, icons, progress, progress_bar
from usdb_syncer.gui.about_dialog import AboutDialog
from usdb_syncer.gui.comment_dialog import CommentDialog
from usdb_syncer.gui.debug_console import DebugConsole
from usdb_syncer.gui.forms.MainWindow import Ui_MainWindow
from usdb_syncer.gui.meta_tags_dialog import MetaTagsDialog
from usdb_syncer.gui.progress import run_with_progress
from usdb_syncer.gui.report_dialog import ReportDialog
from usdb_syncer.gui.search_tree.tree import FilterTree
from usdb_syncer.gui.settings_dialog import SettingsDialog
from usdb_syncer.gui.song_table.song_table import SongTable
from usdb_syncer.gui.usdb_login_dialog import UsdbLoginDialog
from usdb_syncer.logger import logger
from usdb_syncer.song_loader import DownloadManager
from usdb_syncer.sync_meta import SyncMeta
from usdb_syncer.usdb_scraper import post_song_rating
from usdb_syncer.usdb_song import UsdbSong
from usdb_syncer.utils import AppPaths, LinuxEnvCleaner, open_path_or_file


class MainWindow(Ui_MainWindow, QMainWindow):
    """The app's main window and entry point to the GUI."""

    _cleaned_up = False

    def __init__(self) -> None:
        super().__init__()
        self.setupUi(self)
        self.tree = FilterTree(self)
        self.table = SongTable(self)
        self.progress_bar = progress_bar.ProgressBar(
            self.bar_download_progress, self.label_download_progress
        )
        self._setup_statusbar()
        self._setup_log()
        self._setup_toolbar()
        self._setup_shortcuts()
        self._setup_song_dir()
        self.lineEdit_search.textChanged.connect(
            lambda txt: events.TextFilterChanged(txt).post()
        )
        events.SavedSearchRestored.subscribe(
            lambda event: self.lineEdit_search.setText(event.search.text)
        )
        events.ThemeChanged.subscribe(self._on_theme_changed)
        self._setup_buttons()
        self._restore_state()

    def _setup_statusbar(self) -> None:
        self._status_label = QLabel(self)
        self.statusbar.addWidget(self._status_label)

        def on_count_changed(rows: int, selected: int) -> None:
            total = db.usdb_song_count()
            self._status_label.setText(
                f"{rows} out of {total} songs shown, {selected} selected."
            )

        self.table.connect_row_count_changed(on_count_changed)

    def _setup_log(self) -> None:
        self.plainTextEdit.setReadOnly(True)
        self._debugs: list[tuple[str, float]] = []
        self._infos: list[tuple[str, float]] = []
        self._warnings: list[tuple[str, float]] = []
        self._errors: list[tuple[str, float]] = []
        self.toolButton_debugs.toggled.connect(self._on_log_filter_changed)
        self.toolButton_infos.toggled.connect(self._on_log_filter_changed)
        self.toolButton_warnings.toggled.connect(self._on_log_filter_changed)
        self.toolButton_errors.toggled.connect(self._on_log_filter_changed)

    def _setup_toolbar(self) -> None:
        self.menu_view.addAction(self.dock_search.toggleViewAction())
        self.menu_view.addAction(self.dock_log.toggleViewAction())
        for action, func in (
            (self.action_songs_download, self.table.download_selection),
            (self.action_songs_abort, self.table.abort_selected_downloads),
            (self.action_find_local_songs, self._select_local_songs),
            (self.action_refetch_song_list, self._refetch_song_list),
            (self.action_usdb_login, lambda: UsdbLoginDialog(self).show()),
            (self.action_meta_tags, lambda: MetaTagsDialog(self).show()),
            (
                self.action_settings,
                lambda: SettingsDialog(self, self.table.current_song()).show(),
            ),
            (self.action_about, lambda: AboutDialog(self).show()),
            (
                self.action_generate_song_list,
                lambda: ReportDialog(self, self.table).show(),
            ),
            (self.action_import_usdb_ids, self._import_usdb_ids_from_files),
            (self.action_export_usdb_ids, self._export_usdb_ids_to_file),
            (self.action_show_log, lambda: open_path_or_file(AppPaths.log.parent)),
            (self.action_show_in_usdb, self._show_current_song_in_usdb),
            (self.action_post_comment_in_usdb, self._show_comment_dialog),
            (self.action_rate_1star, lambda: self._rate_in_usdb(1)),
            (self.action_rate_2stars, lambda: self._rate_in_usdb(2)),
            (self.action_rate_3stars, lambda: self._rate_in_usdb(3)),
            (self.action_rate_4stars, lambda: self._rate_in_usdb(4)),
            (self.action_rate_5stars, lambda: self._rate_in_usdb(5)),
            (self.action_open_song_folder, self._open_current_song_folder),
            (
                self.action_open_song_in_karedi,
                lambda: self._open_current_song_in_app(settings.SupportedApps.KAREDI),
            ),
            (
                self.action_open_song_in_performous,
                lambda: self._open_current_song_in_app(
                    settings.SupportedApps.PERFORMOUS
                ),
            ),
            (
                self.action_open_song_in_ultrastar_manager,
                lambda: self._open_current_song_in_app(
                    settings.SupportedApps.ULTRASTAR_MANAGER
                ),
            ),
            (
                self.action_open_song_in_usdx,
                lambda: self._open_current_song_in_app(settings.SupportedApps.USDX),
            ),
            (
                self.action_open_song_in_vocaluxe,
                lambda: self._open_current_song_in_app(settings.SupportedApps.VOCALUXE),
            ),
            (
                self.action_open_song_in_yass_reloaded,
                lambda: self._open_current_song_in_app(
                    settings.SupportedApps.YASS_RELOADED
                ),
            ),
            (self.action_delete, self.table.delete_selected_songs),
            (self.action_pin, self.table.set_pin_selected_songs),
        ):
            action.triggered.connect(func)
        self.menu_custom_data.aboutToShow.connect(self.table.build_custom_data_menu)

    def _setup_shortcuts(self) -> None:
        gui_utils.set_shortcut("Ctrl+.", self, lambda: DebugConsole(self).show())

    def _setup_song_dir(self) -> None:
        self.song_dir = settings.get_song_dir()
        self.lineEdit_song_dir.setText(str(self.song_dir))
        self.pushButton_select_song_dir.clicked.connect(self._select_song_dir)

    def _setup_buttons(self) -> None:
        self.button_download.clicked.connect(self.table.download_selection)
        self.button_pause.clicked.connect(DownloadManager.set_pause)

    def _on_log_filter_changed(self) -> None:
        messages = []
        if self.toolButton_debugs.isChecked():
            messages += self._debugs
        if self.toolButton_infos.isChecked():
            messages += self._infos
        if self.toolButton_warnings.isChecked():
            messages += self._warnings
        if self.toolButton_errors.isChecked():
            messages += self._errors
        messages.sort(key=lambda m: m[1])
        self.plainTextEdit.setPlainText("\n".join(m[0] for m in messages))
        gui_utils.scroll_to_bottom(self.plainTextEdit)

    def log_to_text_edit(self, message: str, level: int, created: float) -> None:
        match level:
            case 40:
                self._errors.append((message, created))
                if self.toolButton_errors.isChecked():
                    self.plainTextEdit.appendPlainText(message)
            case 30:
                self._warnings.append((message, created))
                if self.toolButton_warnings.isChecked():
                    self.plainTextEdit.appendPlainText(message)
            case 20:
                self._infos.append((message, created))
                if self.toolButton_infos.isChecked():
                    self.plainTextEdit.appendPlainText(message)
            case 10:
                self._debugs.append((message, created))
                if self.toolButton_debugs.isChecked():
                    self.plainTextEdit.appendPlainText(message)

    def _select_local_songs(self) -> None:
        def on_done(result: progress.Result[set[SongId]]) -> None:
            songs = result.result()
            self.table.set_selection_to_song_ids(songs)
            logger.info(f"Selected {len(songs)} songs.")

        if directory := QFileDialog.getExistingDirectory(self, "Select Song Directory"):
            run_with_progress(
                "Reading song txts ...",
                lambda: song_routines.find_local_songs(Path(directory)),
                on_done=on_done,
            )

    def _refetch_song_list(self) -> None:
        def task() -> None:
            with db.transaction():
                song_routines.load_available_songs(force_reload=True)

        def on_done(result: progress.Result[None]) -> None:
            UsdbSong.clear_cache()
            self.table.end_reset()
            self.table.search_songs()
            result.result()

        self.table.begin_reset()
        run_with_progress("Fetching song list ...", task=task, on_done=on_done)

    def _select_song_dir(self) -> None:
        song_dir = QFileDialog.getExistingDirectory(self, "Select Song Directory")
        if not song_dir:
            return
        path = Path(song_dir).resolve(strict=True)

        def task() -> None:
            with db.transaction():
                song_routines.synchronize_sync_meta_folder(path)
                SyncMeta.reset_active(path)

        def on_done(result: progress.Result[None]) -> None:
            result.result()
            self.lineEdit_song_dir.setText(str(path))
            settings.set_song_dir(path)
            UsdbSong.clear_cache()
            events.SongDirChanged(path).post()

        run_with_progress("Reading meta files ...", task=task, on_done=on_done)

    def _import_usdb_ids_from_files(self) -> None:
        file_list = QFileDialog.getOpenFileNames(
            self,
            caption="Select one or more files to import USDB IDs from",
            dir=str(Path.cwd()),
            filter=(
                "JSON, USDB IDs, Weblinks (*.json *.usdb_ids *.url *.webloc *.desktop)"
            ),
        )[0]
        if not file_list:
            logger.info("no files selected to import USDB IDs from")
            return
        paths = [Path(f) for f in file_list]
        if available := usdb_id_file.get_available_song_ids_from_files(paths):
            self.table.set_selection_to_song_ids(available)

    def _export_usdb_ids_to_file(self) -> None:
        selected_ids = [song.song_id for song in self.table.selected_songs()]
        if not selected_ids:
            logger.info("Skipping export: no songs selected.")
            return

        # Note: automatically checks if file already exists
        path = QFileDialog.getSaveFileName(
            self,
            caption="Select export file for USDB IDs",
            dir=str(Path.cwd()),
            filter="USDB ID File (*.usdb_ids)",
        )[0]
        if not path:
            logger.info("export aborted")
            return

        usdb_id_file.write_usdb_id_file(Path(path), selected_ids)
        logger.info(f"exported {len(selected_ids)} USDB IDs to {path}")

    def _show_current_song_in_usdb(self) -> None:
        if song := self.table.current_song():
            logger.debug(f"Opening song page #{song.song_id} in webbrowser.")
            with LinuxEnvCleaner():
                webbrowser.open(f"{Usdb.DETAIL_URL}{song.song_id:d}")
        else:
            logger.info("No current song.")

    def _show_comment_dialog(self) -> None:
        song = self.table.current_song()
        if song:
            CommentDialog(self, song).show()
        else:
            logger.debug("Not opening comment dialog: no song selected.")

    def _rate_in_usdb(self, stars: int) -> None:
        song = self.table.current_song()
        if song:
            post_song_rating(song.song_id, stars)
        else:
            logger.debug("Not rating song: no song selected.")

    def _open_current_song(self, action: Callable[[Path], None]) -> None:
        if song := self.table.current_song():
            if song.sync_meta:
                if song.sync_meta.path.exists():
                    action(song.sync_meta.path.parent)
                else:
                    with db.transaction():
                        song.remove_sync_meta()
                    events.SongChanged(song.song_id)
                    logger.info("Song does not exist locally anymore.")
            else:
                logger.info("Song does not exist locally.")
        else:
            logger.info("No current song.")

    def _open_current_song_folder(self) -> None:
        self._open_current_song(open_path_or_file)

    def _open_current_song_in_app(self, app: settings.SupportedApps) -> None:
        self._open_current_song(lambda path: settings.SupportedApps.open_app(app, path))

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:  # noqa: N802
        def on_done(result: progress.Result) -> None:
            result.log_error()
            self.table.save_state()
            self._save_state()
            db.close()
            self._cleaned_up = True
            logger.debug("Closing after cleanup.")
            self.close()

        if self._cleaned_up:
            logger.debug("Accepting close event.")
            event.accept()
        else:
            logger.debug("Close event deferred, cleaning up ...")
            run_with_progress("Shutting down ...", DownloadManager.quit, on_done)
            event.ignore()

    def _restore_state(self) -> None:
        self.restoreGeometry(settings.get_geometry_main_window())
        self.restoreState(settings.get_state_main_window())
        self.dock_log.restoreGeometry(settings.get_geometry_log_dock())

    def _save_state(self) -> None:
        settings.set_geometry_main_window(self.saveGeometry())
        settings.set_state_main_window(self.saveState())
        settings.set_geometry_log_dock(self.dock_log.saveGeometry())

    def _on_theme_changed(self, event: events.ThemeChanged) -> None:
        theme = event.theme
        self.toolButton_debugs.setIcon(icons.Icon.BUG.icon(theme))
        self.toolButton_infos.setIcon(icons.Icon.INFO.icon(theme))
        self.toolButton_warnings.setIcon(icons.Icon.WARNING.icon(theme))
        self.toolButton_errors.setIcon(icons.Icon.ERROR.icon(theme))
        self.button_download.setIcon(icons.Icon.DOWNLOAD.icon(theme))
        self.button_pause.setIcon(icons.Icon.PAUSE_REMOTE.icon(theme))
        self.pushButton_select_song_dir.setIcon(icons.Icon.SONG_FOLDER.icon(theme))
        self.action_settings.setIcon(icons.Icon.SETTINGS.icon(theme))
        self.action_meta_tags.setIcon(icons.Icon.META_TAGS.icon(theme))
        self.action_generate_song_list.setIcon(icons.Icon.REPORT.icon(theme))
        self.action_usdb_login.setIcon(icons.Icon.USDB.icon(theme))
        self.action_refetch_song_list.setIcon(icons.Icon.CHECK_FOR_UPDATE.icon(theme))
        self.action_show_log.setIcon(icons.Icon.LOG.icon(theme))
        self.action_songs_download.setIcon(icons.Icon.DOWNLOAD.icon(theme))
        self.action_songs_abort.setIcon(icons.Icon.ABORT.icon(theme))
        self.action_show_in_usdb.setIcon(icons.Icon.USDB.icon(theme))
        self.action_post_comment_in_usdb.setIcon(icons.Icon.COMMENT.icon(theme))
        self.menu_rate_song_on_usdb.setIcon(icons.Icon.RATING.icon(theme))
        self.action_open_song_folder.setIcon(icons.Icon.SONG_FOLDER.icon(theme))
        self.menu_open_song_in.setIcon(icons.Icon.OPEN_SONG_WITH.icon(theme))
        self.menu_custom_data.setIcon(icons.Icon.CUSTOM_DATA.icon(theme))
        self.action_pin.setIcon(icons.Icon.PIN.icon(theme))
        self.action_delete.setIcon(icons.Icon.DELETE.icon(theme))
        self.action_open_song_in_usdx.setIcon(icons.Icon.USDX.icon(theme))
        self.action_open_song_in_vocaluxe.setIcon(icons.Icon.VOCALUXE.icon(theme))
        self.action_open_song_in_performous.setIcon(icons.Icon.PERFORMOUS.icon(theme))
        self.action_open_song_in_yass_reloaded.setIcon(
            icons.Icon.YASS_RELOADED.icon(theme)
        )
        self.action_open_song_in_karedi.setIcon(icons.Icon.KAREDI.icon(theme))
        self.action_open_song_in_ultrastar_manager.setIcon(
            icons.Icon.ULTRASTAR_MANAGER.icon(theme)
        )
        self.action_find_local_songs.setIcon(icons.Icon.DATABASE.icon(theme))
        self.action_import_usdb_ids.setIcon(icons.Icon.FILE_IMPORT.icon(theme))
        self.action_export_usdb_ids.setIcon(icons.Icon.FILE_EXPORT.icon(theme))



"""Manages the media player singleton."""

from PySide6 import QtMultimedia


class _MediaPlayer:
    player: QtMultimedia.QMediaPlayer | None = None


def media_player() -> QtMultimedia.QMediaPlayer:
    if _MediaPlayer.player is None:
        _MediaPlayer.player = QtMultimedia.QMediaPlayer()
        _MediaPlayer.player.setAudioOutput(
            QtMultimedia.QAudioOutput(_MediaPlayer.player)
        )
    return _MediaPlayer.player



"""Meta data about the synchronization state of a USDB song."""

from __future__ import annotations

import json
from collections.abc import Iterator
from pathlib import Path
from typing import Any

import attrs

from usdb_syncer import SongId, SyncMetaId, db, settings, utils
from usdb_syncer.custom_data import CustomData
from usdb_syncer.logger import logger
from usdb_syncer.meta_tags import MetaTags

SYNC_META_VERSION = 1
# mtimes may deviate up to 2 seconds on different file systems
# See https://en.wikipedia.org/wiki/File_Allocation_Table
MTIME_TOLERANCE_SECS = 2


class SyncMetaTooNewError(Exception):
    """Raised when trying to decode meta info from an incompatible future release."""

    def __str__(self) -> str:
        return "cannot read sync meta written by a future release"


@attrs.define
class ResourceFile:
    """Meta data about a local file."""

    fname: str
    mtime: int
    resource: str

    @classmethod
    def new(cls, path: Path, resource: str) -> ResourceFile:
        return cls(path.name, utils.get_mtime(path), resource)

    @classmethod
    def from_nested_dict(cls, dct: Any) -> ResourceFile | None:
        if (
            isinstance(dct, dict)
            and isinstance(fname := dct.get("fname"), str)
            and isinstance(mtime := dct.get("mtime"), (int, float))
            and isinstance(resource := dct.get("resource"), str)
        ):
            return cls(fname=fname, mtime=int(mtime), resource=resource)
        return None

    @classmethod
    def from_db_row(
        cls, row: tuple[str | None, int | None, str | None]
    ) -> ResourceFile | None:
        if row[0] is None or row[1] is None or row[2] is None:
            return None
        return cls(fname=row[0], mtime=row[1], resource=row[2])

    def is_in_sync(self, folder: Path) -> bool:
        """True if this file exists in the given folder and is in sync."""
        path = folder.joinpath(self.fname)
        return (
            path.exists()
            and abs(utils.get_mtime(path) - self.mtime) / 1_000_000
            < MTIME_TOLERANCE_SECS
        )

    def db_params(
        self, sync_meta_id: SyncMetaId, kind: db.ResourceFileKind
    ) -> db.ResourceFileParams:
        return db.ResourceFileParams(
            sync_meta_id=sync_meta_id,
            kind=kind,
            fname=self.fname,
            mtime=self.mtime,
            resource=self.resource,
        )


@attrs.define
class SyncMeta:
    """Meta data about the synchronization state of a USDB song."""

    sync_meta_id: SyncMetaId
    song_id: SongId
    path: Path
    mtime: int
    meta_tags: MetaTags
    pinned: bool = False
    txt: ResourceFile | None = None
    audio: ResourceFile | None = None
    video: ResourceFile | None = None
    cover: ResourceFile | None = None
    background: ResourceFile | None = None
    custom_data: CustomData = attrs.field(factory=CustomData)

    @classmethod
    def new(cls, song_id: SongId, folder: Path, meta_tags: MetaTags) -> SyncMeta:
        sync_meta_id = SyncMetaId.new()
        return cls(
            sync_meta_id=sync_meta_id,
            song_id=song_id,
            path=folder.joinpath(sync_meta_id.to_filename()),
            mtime=0,
            meta_tags=meta_tags,
        )

    @classmethod
    def try_from_file(cls, path: Path) -> SyncMeta | None:
        new_id = False
        if (sync_meta_id := SyncMetaId.from_path(path)) is None:
            # might be a legacy file with old-style id
            sync_meta_id = SyncMetaId.new()
            new_id = True
        with path.open(encoding="utf8") as file:
            try:
                dct = json.load(file)
            except (json.decoder.JSONDecodeError, UnicodeDecodeError):
                return None
        if not isinstance(dct, dict):
            return None
        if int(dct["version"]) > SYNC_META_VERSION:
            raise SyncMetaTooNewError
        try:
            meta = cls(
                sync_meta_id=sync_meta_id,
                song_id=SongId(dct["song_id"]),
                path=path,
                mtime=utils.get_mtime(path),
                meta_tags=MetaTags.parse(dct["meta_tags"], logger),
                pinned=bool(dct.get("pinned", False)),
                txt=ResourceFile.from_nested_dict(dct["txt"]),
                audio=ResourceFile.from_nested_dict(dct["audio"]),
                video=ResourceFile.from_nested_dict(dct["video"]),
                cover=ResourceFile.from_nested_dict(dct["cover"]),
                background=ResourceFile.from_nested_dict(dct["background"]),
                custom_data=CustomData(dct.get("custom_data")),
            )
        except (TypeError, KeyError, ValueError):
            return None
        if new_id:
            meta.path = path.with_name(sync_meta_id.to_filename())
            path.rename(meta.path)
            logger.info(f"Assigned new ID to meta file: '{path}' > '{meta.path}'.")
        return meta

    @classmethod
    def from_db_row(cls, row: tuple) -> SyncMeta:
        assert len(row) == 21
        meta = cls(
            sync_meta_id=SyncMetaId(row[0]),
            song_id=SongId(row[1]),
            path=Path(row[2]),
            mtime=row[3],
            meta_tags=MetaTags.parse(row[4], logger),
            pinned=bool(row[5]),
        )
        meta.txt = ResourceFile.from_db_row(row[6:9])
        meta.audio = ResourceFile.from_db_row(row[9:12])
        meta.video = ResourceFile.from_db_row(row[12:15])
        meta.cover = ResourceFile.from_db_row(row[15:18])
        meta.background = ResourceFile.from_db_row(row[18:])
        meta.custom_data = CustomData(db.get_custom_data(meta.sync_meta_id))
        return meta

    @classmethod
    def get_in_folder(cls, folder: Path) -> Iterator[SyncMeta]:
        return (SyncMeta.from_db_row(r) for r in db.get_in_folder(folder))

    @classmethod
    def reset_active(cls, folder: Path) -> None:
        db.reset_active_sync_metas(folder)

    def upsert(self) -> None:
        db.upsert_sync_meta(self.db_params())
        db.update_active_sync_metas(settings.get_song_dir(), self.song_id)
        files = self.all_resource_files()
        db.upsert_resource_files(
            file.db_params(self.sync_meta_id, kind) for file, kind in files if file
        )
        db.delete_resource_files(
            (self.sync_meta_id, kind) for file, kind in files if not file
        )
        db.delete_custom_meta_data((self.sync_meta_id,))
        db.upsert_custom_meta_data(
            db.CustomMetaDataParams(self.sync_meta_id, k, v)
            for k, v in self.custom_data.items()
        )

    @classmethod
    def upsert_many(cls, metas: list[SyncMeta]) -> None:
        db.upsert_sync_metas(meta.db_params() for meta in metas)
        db.reset_active_sync_metas(settings.get_song_dir())
        db.upsert_resource_files(
            file.db_params(meta.sync_meta_id, kind)
            for meta in metas
            for file, kind in meta.all_resource_files()
            if file
        )
        db.delete_resource_files(
            (meta.sync_meta_id, kind)
            for meta in metas
            for file, kind in meta.all_resource_files()
            if not file
        )
        db.delete_custom_meta_data(m.sync_meta_id for m in metas)
        db.upsert_custom_meta_data(
            db.CustomMetaDataParams(m.sync_meta_id, k, v)
            for m in metas
            for k, v in m.custom_data.items()
        )

    def delete(self) -> None:
        db.delete_sync_meta(self.sync_meta_id)

    @classmethod
    def delete_many(cls, ids: tuple[SyncMetaId, ...]) -> None:
        db.delete_sync_metas(ids)

    def all_resource_files(
        self,
    ) -> tuple[tuple[ResourceFile | None, db.ResourceFileKind], ...]:
        return (
            (self.txt, db.ResourceFileKind.TXT),
            (self.audio, db.ResourceFileKind.AUDIO),
            (self.video, db.ResourceFileKind.VIDEO),
            (self.cover, db.ResourceFileKind.COVER),
            (self.background, db.ResourceFileKind.BACKGROUND),
        )

    def db_params(self) -> db.SyncMetaParams:
        return db.SyncMetaParams(
            sync_meta_id=self.sync_meta_id,
            song_id=self.song_id,
            path=self.path.as_posix(),
            mtime=self.mtime,
            meta_tags=str(self.meta_tags),
            pinned=self.pinned,
        )

    def synchronize_to_file(self) -> None:
        """Rewrite the file on disk and update the mtime."""
        with self.path.open("w", encoding="utf8") as file:
            json.dump(self, file, cls=SyncMetaEncoder)
        self.mtime = utils.get_mtime(self.path)

    def resource_files(self) -> Iterator[ResourceFile]:
        for meta in (self.txt, self.audio, self.video, self.cover, self.background):
            if meta:
                yield meta


class SyncMetaEncoder(json.JSONEncoder):
    """Custom JSON encoder"""

    def default(self, o: Any) -> Any:
        if isinstance(o, ResourceFile):
            return attrs.asdict(o)
        if isinstance(o, MetaTags):
            return str(o)
        if isinstance(o, SyncMeta):
            fields = attrs.fields(SyncMeta)
            filt = attrs.filters.exclude(fields.sync_meta_id, fields.path, fields.mtime)
            dct = attrs.asdict(o, recurse=False, filter=filt)
            dct["version"] = SYNC_META_VERSION
            return dct
        if isinstance(o, CustomData):
            return o.inner()
        return super().default(o)



"""Dialog to create meta tags."""

import re
from typing import Literal

from PySide6.QtGui import QGuiApplication
from PySide6.QtWidgets import QDialog, QWidget

from usdb_syncer.gui.forms.MetaTagsDialog import Ui_Dialog
from usdb_syncer.meta_tags import (
    CropMetaTags,
    ImageMetaTags,
    MedleyTag,
    MetaTags,
    ResizeMetaTags,
)
from usdb_syncer.utils import extract_vimeo_id, extract_youtube_id

URL_TAG_NAMES = ("cover_url", "background_url", "audio_url", "video_url")


class MetaTagsDialog(Ui_Dialog, QDialog):
    """Dialog to create meta tags."""

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self.setupUi(self)
        self._connect_signals()
        self._shortened_urls: dict[str, str] = {}

    def _connect_signals(self) -> None:
        for signal in (
            self.audio_url.textChanged,
            self.video_url.textChanged,
            self.cover_url.textChanged,
            self.cover_rotation.valueChanged,
            self.cover_contrast.valueChanged,
            self.cover_contrast_auto.toggled,
            self.cover_contrast_auto.toggled,
            self.cover_resize.valueChanged,
            self.cover_crop_left.valueChanged,
            self.cover_crop_top.valueChanged,
            self.cover_crop_width.valueChanged,
            self.cover_crop_height.valueChanged,
            self.background_url.textChanged,
            self.background_resize_width.valueChanged,
            self.background_resize_height.valueChanged,
            self.background_crop_left.valueChanged,
            self.background_crop_top.valueChanged,
            self.background_crop_width.valueChanged,
            self.background_crop_height.valueChanged,
            self.duet.toggled,
            self.duet_p1.textChanged,
            self.duet_p2.textChanged,
            self.preview_start.valueChanged,
            self.medley_start.valueChanged,
            self.medley_end.valueChanged,
            self.tags.textChanged,
        ):
            signal.connect(lambda: self.output.setText(f"#VIDEO:{self._meta_tags()}"))

        self.button_copy_to_clipboard.pressed.connect(self._on_copy_to_clipboard)

    def _audio_source(self) -> str | None:
        if not (source := self.audio_url.text()) or source == self.video_url.text():
            return None
        return _sanitize_video_url(source)

    def _cover_meta_tags(self) -> ImageMetaTags | None:
        if not (cover_source := _sanitize_image_url(self.cover_url.text())):
            return None
        return ImageMetaTags(
            source=cover_source,
            rotate=round(self.cover_rotation.value(), 2) or None,
            crop=self._cover_crop_meta_tag(),
            resize=self._cover_resize_meta_tag(),
            contrast=self._contrast(),
        )

    def _contrast(self) -> Literal["auto"] | float | None:
        if self.cover_contrast_auto.isChecked():
            return "auto"
        if (contrast := round(self.cover_contrast.value(), 2)) == 1:
            return None
        return contrast

    def _cover_crop_meta_tag(self) -> CropMetaTags | None:
        if not (width := self.cover_crop_width.value()) or not (
            height := self.cover_crop_height.value()
        ):
            return None
        left = self.cover_crop_left.value()
        upper = self.cover_crop_top.value()
        return CropMetaTags(
            left=left, upper=upper, right=left + width, lower=upper + height
        )

    def _cover_resize_meta_tag(self) -> ResizeMetaTags | None:
        if resize_value := self.cover_resize.value():
            return ResizeMetaTags(resize_value, resize_value)
        return None

    def _background_meta_tags(self) -> ImageMetaTags | None:
        if not (bg_source := _sanitize_image_url(self.background_url.text())):
            return None
        return ImageMetaTags(
            source=bg_source,
            crop=self._background_crop_meta_tag(),
            resize=self._background_resize_meta_tag(),
        )

    def _background_crop_meta_tag(self) -> CropMetaTags | None:
        if not (width := self.background_crop_width.value()) or not (
            height := self.background_crop_height.value()
        ):
            return None
        left = self.background_crop_left.value()
        upper = self.background_crop_top.value()
        return CropMetaTags(
            left=left, upper=upper, right=left + width, lower=upper + height
        )

    def _background_resize_meta_tag(self) -> ResizeMetaTags | None:
        if not (width := self.background_resize_width.value()) or not (
            height := self.background_resize_height.value()
        ):
            return None
        return ResizeMetaTags(width, height)

    def _medley_tag(self) -> MedleyTag | None:
        if (start := self.medley_start.value()) < (end := self.medley_end.value()):
            return MedleyTag(start, end)
        return None

    def _p1_meta_tag(self) -> str | None:
        if self.duet.isChecked():
            return self.duet_p1.text() or "P1"
        return None

    def _p2_meta_tag(self) -> str | None:
        if self.duet.isChecked():
            return self.duet_p2.text() or "P2"
        return None

    def _meta_tags(self) -> MetaTags:
        return MetaTags(
            audio=self._audio_source(),
            video=_sanitize_video_url(self.video_url.text()) or None,
            cover=self._cover_meta_tags(),
            background=self._background_meta_tags(),
            player1=self._p1_meta_tag(),
            player2=self._p2_meta_tag(),
            preview=round(self.preview_start.value(), 3) or None,
            medley=self._medley_tag(),
            tags=self.tags.text() or None,
        )

    def _toggle_auto_contrast(self) -> None:
        auto = self.cover_contrast_auto.isChecked()
        self.cover_contrast.setEnabled(not auto)

    def _on_copy_to_clipboard(self) -> None:
        QGuiApplication.clipboard().setText(self.output.text())


def _sanitize_video_url(url: str) -> str:
    """Returns a YouTube id, Vimeo id or sanitized URL."""
    return extract_youtube_id(url) or extract_vimeo_id(url) or url


def _sanitize_image_url(url: str) -> str:
    """Returns a fanart id or sanitized URL and whether it uses HTTP."""
    if "fanart.tv" in url:
        return url.removeprefix("https://images.fanart.tv/fanart/")
    if "m.media-amazon.com" in url:
        pattern = r"\._.*[^_\s]_"
        return re.sub(pattern, "", url)

    return url



"""Custom theming inspired by Material Design."""

from __future__ import annotations

import enum
from importlib import resources
from typing import assert_never

import attrs
from PySide6 import QtWidgets
from PySide6.QtGui import QColor, QColorConstants, QPalette

from usdb_syncer import events, settings
from usdb_syncer.gui.resources import styles


def apply_theme(
    theme: settings.Theme, primary_color: settings.Color, colored_background: bool
) -> None:
    app = QtWidgets.QApplication.instance()
    assert isinstance(app, QtWidgets.QApplication), "Theming requires a GUI application"
    match theme:
        case settings.Theme.SYSTEM:
            q_palette = QPalette()
            style = ""
        case settings.Theme.DARK:
            dark_theme = DarkTheme(primary_color, colored_background)
            q_palette = dark_theme.q_palette()
            style = dark_theme.style()
        case unreachable:
            assert_never(unreachable)
    app.setPalette(q_palette)
    app.setStyleSheet(style)
    events.ThemeChanged(theme).post()


class _Surface(enum.Enum):
    """A surface of a certain level."""

    DP_00 = "surface_00dp", 0.0
    DP_01 = "surface_01dp", 0.05
    DP_02 = "surface_02dp", 0.07
    DP_03 = "surface_03dp", 0.08
    DP_04 = "surface_04dp", 0.09
    DP_06 = "surface_06dp", 0.11
    DP_08 = "surface_08dp", 0.12
    DP_12 = "surface_12dp", 0.14
    DP_16 = "surface_16dp", 0.15
    DP_24 = "surface_24dp", 0.16

    def __str__(self) -> str:
        return self.value[0]

    def overlay_alpha(self) -> float:
        return self.value[1]


class _Text(enum.Enum):
    """Text of a certain emphasis."""

    HIGH = "text_high", 0.87
    MEDIUM = "text_medium", 0.60
    DISABLED = "text_disabled", 0.38

    def __str__(self) -> str:
        return self.value[0]

    def opacity(self) -> float:
        return self.value[1]


def _overlay_colors(color_1: QColor, color_2: QColor, alpha: float) -> QColor:
    return QColor(
        _weighted_sum(color_1.red(), color_2.red(), alpha),
        _weighted_sum(color_1.green(), color_2.green(), alpha),
        _weighted_sum(color_1.blue(), color_2.blue(), alpha),
    )


def _weighted_sum(x: int, y: int, y_weight: float) -> int:
    return round((1 - y_weight) * x + y_weight * y)


def _rgb_str(color: QColor) -> str:
    return f"rgb({color.red()}, {color.green()}, {color.blue()})"


@attrs.define
class DarkTheme:
    """Style and palette builder for a dark theme."""

    primary_color: settings.Color
    colored_background: bool
    on_primary: QColor = QColorConstants.Svg.black
    on_secondary: QColor = QColorConstants.Svg.black
    on_background: QColor = QColorConstants.Svg.white
    on_surface: QColor = QColorConstants.Svg.white
    _style_template = resources.files(styles).joinpath("dark.qss")
    primary_swatch: Swatch = attrs.field(init=False)
    primary: QColor = attrs.field(init=False)
    base_surface: QColor = attrs.field(init=False)
    default_surface: QColor = attrs.field(init=False, default=QColor(18, 18, 18))

    def __attrs_post_init__(self) -> None:
        self.primary_swatch = Swatch.get(self.primary_color)
        self.primary = self.primary_swatch.s_200
        if self.colored_background:
            self.base_surface = _overlay_colors(
                self.default_surface, self.primary_swatch.s_900, 0.08
            )
        else:
            self.base_surface = self.default_surface

    def surface(self, variant: _Surface) -> QColor:
        return _overlay_colors(
            self.base_surface, QColorConstants.Svg.white, variant.overlay_alpha()
        )

    def text(self, variant: _Text) -> QColor:
        return QColor(
            round(self.on_background.red() * variant.opacity()),
            round(self.on_background.green() * variant.opacity()),
            round(self.on_background.blue() * variant.opacity()),
        )

    def q_palette(self) -> QPalette:
        palette = QPalette()
        high_text = self.text(_Text.HIGH)
        palette.setColor(QPalette.ColorRole.WindowText, high_text)
        palette.setColor(QPalette.ColorRole.Button, self.surface(_Surface.DP_06))
        palette.setColor(QPalette.ColorRole.Text, high_text)
        palette.setColor(QPalette.ColorRole.BrightText, self.primary)
        palette.setColor(QPalette.ColorRole.ButtonText, high_text)
        palette.setColor(QPalette.ColorRole.Base, self.surface(_Surface.DP_01))
        palette.setColor(QPalette.ColorRole.Window, self.base_surface)
        palette.setColor(QPalette.ColorRole.Highlight, self.primary)
        palette.setColor(QPalette.ColorRole.HighlightedText, self.on_primary)
        palette.setColor(QPalette.ColorRole.Link, Swatch.get(settings.Color.BLUE).s_200)
        palette.setColor(
            QPalette.ColorRole.LinkVisited, Swatch.get(settings.Color.PURPLE).s_200
        )
        palette.setColor(QPalette.ColorRole.AlternateBase, self.surface(_Surface.DP_02))
        palette.setColor(QPalette.ColorRole.ToolTipBase, self.primary)
        palette.setColor(QPalette.ColorRole.ToolTipText, self.on_primary)
        palette.setColor(QPalette.ColorRole.PlaceholderText, self.text(_Text.DISABLED))
        palette.setColor(QPalette.ColorRole.Accent, self.primary)
        return palette

    def style(self) -> str:
        surfaces = {str(s): _rgb_str(self.surface(s)) for s in _Surface}
        texts = {str(t): _rgb_str(self.text(t)) for t in _Text}
        return self._style_template.read_text(encoding="utf-8").format(
            **surfaces,
            **texts,
            primary=_rgb_str(self.primary_swatch.s_200),
            on_primary=_rgb_str(self.on_primary),
            on_secondary=_rgb_str(self.on_secondary),
            on_background=_rgb_str(self.on_background),
            on_surface=_rgb_str(self.on_surface),
        )


@attrs.define
class Swatch:
    """A Material Design swatch."""

    s_50: QColor
    s_100: QColor
    s_200: QColor
    s_300: QColor
    s_400: QColor
    s_500: QColor
    s_600: QColor
    s_700: QColor
    s_800: QColor
    s_900: QColor
    s_a100: QColor | None = None
    s_a200: QColor | None = None
    s_a400: QColor | None = None
    s_a700: QColor | None = None

    @classmethod
    def get(cls, color: settings.Color) -> Swatch:  # noqa: C901
        match color:
            case settings.Color.RED:
                return Swatch(
                    s_50=QColor(255, 235, 238),
                    s_100=QColor(255, 205, 210),
                    s_200=QColor(239, 154, 154),
                    s_300=QColor(229, 115, 115),
                    s_400=QColor(239, 83, 80),
                    s_500=QColor(244, 67, 54),
                    s_600=QColor(229, 57, 53),
                    s_700=QColor(211, 47, 47),
                    s_800=QColor(198, 40, 40),
                    s_900=QColor(183, 28, 28),
                    s_a100=QColor(255, 138, 128),
                    s_a200=QColor(255, 82, 82),
                    s_a400=QColor(255, 23, 68),
                    s_a700=QColor(213, 0, 0),
                )
            case settings.Color.PINK:
                return Swatch(
                    s_50=QColor(252, 228, 236),
                    s_100=QColor(248, 187, 208),
                    s_200=QColor(244, 143, 177),
                    s_300=QColor(240, 98, 146),
                    s_400=QColor(236, 64, 122),
                    s_500=QColor(233, 30, 99),
                    s_600=QColor(216, 27, 96),
                    s_700=QColor(194, 24, 91),
                    s_800=QColor(173, 20, 87),
                    s_900=QColor(136, 14, 79),
                    s_a100=QColor(255, 128, 171),
                    s_a200=QColor(255, 64, 129),
                    s_a400=QColor(245, 0, 87),
                    s_a700=QColor(197, 17, 98),
                )
            case settings.Color.PURPLE:
                return Swatch(
                    s_50=QColor(243, 229, 245),
                    s_100=QColor(225, 190, 231),
                    s_200=QColor(206, 147, 216),
                    s_300=QColor(186, 104, 200),
                    s_400=QColor(171, 71, 188),
                    s_500=QColor(156, 39, 176),
                    s_600=QColor(142, 36, 170),
                    s_700=QColor(123, 31, 162),
                    s_800=QColor(106, 27, 154),
                    s_900=QColor(74, 20, 140),
                    s_a100=QColor(234, 128, 252),
                    s_a200=QColor(224, 64, 251),
                    s_a400=QColor(213, 0, 249),
                    s_a700=QColor(170, 0, 255),
                )
            case settings.Color.DEEPPURPLE:
                return Swatch(
                    s_50=QColor(237, 231, 246),
                    s_100=QColor(209, 196, 233),
                    s_200=QColor(179, 157, 219),
                    s_300=QColor(149, 117, 205),
                    s_400=QColor(126, 87, 194),
                    s_500=QColor(103, 58, 183),
                    s_600=QColor(94, 53, 177),
                    s_700=QColor(81, 45, 168),
                    s_800=QColor(69, 39, 160),
                    s_900=QColor(49, 27, 146),
                    s_a100=QColor(179, 136, 255),
                    s_a200=QColor(124, 77, 255),
                    s_a400=QColor(101, 31, 255),
                    s_a700=QColor(98, 0, 234),
                )
            case settings.Color.INDIGO:
                return Swatch(
                    s_50=QColor(232, 234, 246),
                    s_100=QColor(197, 202, 233),
                    s_200=QColor(159, 168, 218),
                    s_300=QColor(121, 134, 203),
                    s_400=QColor(92, 107, 192),
                    s_500=QColor(63, 81, 181),
                    s_600=QColor(57, 73, 171),
                    s_700=QColor(48, 63, 159),
                    s_800=QColor(40, 53, 147),
                    s_900=QColor(26, 35, 126),
                    s_a100=QColor(140, 158, 255),
                    s_a200=QColor(83, 109, 254),
                    s_a400=QColor(61, 90, 254),
                    s_a700=QColor(48, 79, 254),
                )
            case settings.Color.BLUE:
                return Swatch(
                    s_50=QColor(227, 242, 253),
                    s_100=QColor(187, 222, 251),
                    s_200=QColor(144, 202, 249),
                    s_300=QColor(100, 181, 246),
                    s_400=QColor(66, 165, 245),
                    s_500=QColor(33, 150, 243),
                    s_600=QColor(30, 136, 229),
                    s_700=QColor(25, 118, 210),
                    s_800=QColor(21, 101, 192),
                    s_900=QColor(13, 71, 161),
                    s_a100=QColor(130, 177, 255),
                    s_a200=QColor(68, 138, 255),
                    s_a400=QColor(41, 121, 255),
                    s_a700=QColor(41, 98, 255),
                )
            case settings.Color.LIGHTBLUE:
                return Swatch(
                    s_50=QColor(225, 245, 254),
                    s_100=QColor(179, 229, 252),
                    s_200=QColor(129, 212, 250),
                    s_300=QColor(79, 195, 247),
                    s_400=QColor(41, 182, 246),
                    s_500=QColor(3, 169, 244),
                    s_600=QColor(3, 155, 229),
                    s_700=QColor(2, 136, 209),
                    s_800=QColor(2, 119, 189),
                    s_900=QColor(1, 87, 155),
                    s_a100=QColor(128, 216, 255),
                    s_a200=QColor(64, 196, 255),
                    s_a400=QColor(0, 176, 255),
                    s_a700=QColor(0, 145, 234),
                )
            case settings.Color.CYAN:
                return Swatch(
                    s_50=QColor(224, 247, 250),
                    s_100=QColor(178, 235, 242),
                    s_200=QColor(128, 222, 234),
                    s_300=QColor(77, 208, 225),
                    s_400=QColor(38, 198, 218),
                    s_500=QColor(0, 188, 212),
                    s_600=QColor(0, 172, 193),
                    s_700=QColor(0, 151, 167),
                    s_800=QColor(0, 131, 143),
                    s_900=QColor(0, 96, 100),
                    s_a100=QColor(132, 255, 255),
                    s_a200=QColor(24, 255, 255),
                    s_a400=QColor(0, 229, 255),
                    s_a700=QColor(0, 184, 212),
                )
            case settings.Color.TEAL:
                return Swatch(
                    s_50=QColor(224, 242, 241),
                    s_100=QColor(178, 223, 219),
                    s_200=QColor(128, 203, 196),
                    s_300=QColor(77, 182, 172),
                    s_400=QColor(38, 166, 154),
                    s_500=QColor(0, 150, 136),
                    s_600=QColor(0, 137, 123),
                    s_700=QColor(0, 121, 107),
                    s_800=QColor(0, 105, 92),
                    s_900=QColor(0, 77, 64),
                    s_a100=QColor(167, 255, 235),
                    s_a200=QColor(100, 255, 218),
                    s_a400=QColor(29, 233, 182),
                    s_a700=QColor(0, 191, 165),
                )
            case settings.Color.GREEN:
                return Swatch(
                    s_50=QColor(232, 245, 233),
                    s_100=QColor(200, 230, 201),
                    s_200=QColor(165, 214, 167),
                    s_300=QColor(129, 199, 132),
                    s_400=QColor(102, 187, 106),
                    s_500=QColor(76, 175, 80),
                    s_600=QColor(67, 160, 71),
                    s_700=QColor(56, 142, 60),
                    s_800=QColor(46, 125, 50),
                    s_900=QColor(27, 94, 32),
                    s_a100=QColor(185, 246, 202),
                    s_a200=QColor(105, 240, 174),
                    s_a400=QColor(0, 230, 118),
                    s_a700=QColor(0, 200, 83),
                )
            case settings.Color.LIGHTGREEN:
                return Swatch(
                    s_50=QColor(241, 248, 233),
                    s_100=QColor(220, 237, 200),
                    s_200=QColor(197, 225, 165),
                    s_300=QColor(174, 213, 129),
                    s_400=QColor(156, 204, 101),
                    s_500=QColor(139, 195, 74),
                    s_600=QColor(124, 179, 66),
                    s_700=QColor(104, 159, 56),
                    s_800=QColor(85, 139, 47),
                    s_900=QColor(51, 105, 30),
                    s_a100=QColor(204, 255, 144),
                    s_a200=QColor(178, 255, 89),
                    s_a400=QColor(118, 255, 3),
                    s_a700=QColor(100, 221, 23),
                )
            case settings.Color.LIME:
                return Swatch(
                    s_50=QColor(249, 251, 231),
                    s_100=QColor(240, 244, 195),
                    s_200=QColor(230, 238, 156),
                    s_300=QColor(220, 231, 117),
                    s_400=QColor(212, 225, 87),
                    s_500=QColor(205, 220, 57),
                    s_600=QColor(192, 202, 51),
                    s_700=QColor(175, 180, 43),
                    s_800=QColor(158, 157, 36),
                    s_900=QColor(130, 119, 23),
                    s_a100=QColor(244, 255, 129),
                    s_a200=QColor(238, 255, 65),
                    s_a400=QColor(198, 255, 0),
                    s_a700=QColor(174, 234, 0),
                )
            case settings.Color.YELLOW:
                return Swatch(
                    s_50=QColor(255, 253, 231),
                    s_100=QColor(255, 249, 196),
                    s_200=QColor(255, 245, 157),
                    s_300=QColor(255, 241, 118),
                    s_400=QColor(255, 238, 88),
                    s_500=QColor(255, 235, 59),
                    s_600=QColor(253, 216, 53),
                    s_700=QColor(251, 192, 45),
                    s_800=QColor(249, 168, 37),
                    s_900=QColor(245, 127, 23),
                    s_a100=QColor(255, 255, 141),
                    s_a200=QColor(255, 255, 0),
                    s_a400=QColor(255, 234, 0),
                    s_a700=QColor(255, 214, 0),
                )
            case settings.Color.AMBER:
                return Swatch(
                    s_50=QColor(255, 248, 225),
                    s_100=QColor(255, 236, 179),
                    s_200=QColor(255, 224, 130),
                    s_300=QColor(255, 213, 79),
                    s_400=QColor(255, 202, 40),
                    s_500=QColor(255, 193, 7),
                    s_600=QColor(255, 179, 0),
                    s_700=QColor(255, 160, 0),
                    s_800=QColor(255, 143, 0),
                    s_900=QColor(255, 111, 0),
                    s_a100=QColor(255, 229, 127),
                    s_a200=QColor(255, 215, 64),
                    s_a400=QColor(255, 196, 0),
                    s_a700=QColor(255, 171, 0),
                )
            case settings.Color.ORANGE:
                return Swatch(
                    s_50=QColor(255, 243, 224),
                    s_100=QColor(255, 224, 178),
                    s_200=QColor(255, 204, 128),
                    s_300=QColor(255, 183, 77),
                    s_400=QColor(255, 167, 38),
                    s_500=QColor(255, 152, 0),
                    s_600=QColor(251, 140, 0),
                    s_700=QColor(245, 124, 0),
                    s_800=QColor(239, 108, 0),
                    s_900=QColor(230, 81, 0),
                    s_a100=QColor(255, 209, 128),
                    s_a200=QColor(255, 171, 64),
                    s_a400=QColor(255, 145, 0),
                    s_a700=QColor(255, 109, 0),
                )

            case settings.Color.DEEPORANGE:
                return Swatch(
                    s_50=QColor(251, 233, 231),
                    s_100=QColor(255, 204, 188),
                    s_200=QColor(255, 171, 145),
                    s_300=QColor(255, 138, 101),
                    s_400=QColor(255, 112, 67),
                    s_500=QColor(255, 87, 34),
                    s_600=QColor(244, 81, 30),
                    s_700=QColor(230, 74, 25),
                    s_800=QColor(216, 67, 21),
                    s_900=QColor(191, 54, 12),
                    s_a100=QColor(255, 158, 128),
                    s_a200=QColor(255, 110, 64),
                    s_a400=QColor(255, 61, 0),
                    s_a700=QColor(221, 44, 0),
                )

            case settings.Color.BROWN:
                return Swatch(
                    s_50=QColor(239, 235, 233),
                    s_100=QColor(215, 204, 200),
                    s_200=QColor(188, 170, 164),
                    s_300=QColor(161, 136, 127),
                    s_400=QColor(141, 110, 99),
                    s_500=QColor(121, 85, 72),
                    s_600=QColor(109, 76, 65),
                    s_700=QColor(93, 64, 55),
                    s_800=QColor(78, 52, 46),
                    s_900=QColor(62, 39, 35),
                )
            case settings.Color.GRAY:
                return Swatch(
                    s_50=QColor(250, 250, 250),
                    s_100=QColor(245, 245, 245),
                    s_200=QColor(238, 238, 238),
                    s_300=QColor(224, 224, 224),
                    s_400=QColor(189, 189, 189),
                    s_500=QColor(158, 158, 158),
                    s_600=QColor(117, 117, 117),
                    s_700=QColor(97, 97, 97),
                    s_800=QColor(66, 66, 66),
                    s_900=QColor(33, 33, 33),
                )
            case settings.Color.BLUEGRAY:
                return Swatch(
                    s_50=QColor(236, 239, 241),
                    s_100=QColor(207, 216, 220),
                    s_200=QColor(176, 190, 197),
                    s_300=QColor(144, 164, 174),
                    s_400=QColor(120, 144, 156),
                    s_500=QColor(96, 125, 139),
                    s_600=QColor(84, 110, 122),
                    s_700=QColor(69, 90, 100),
                    s_800=QColor(55, 71, 79),
                    s_900=QColor(38, 50, 56),
                )



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'AboutDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QDialog, QFrame, QLabel,
    QSizePolicy, QTextBrowser, QWidget)
from usdb_syncer.gui.resources.qt import resources as resources_rc

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(500, 300)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setMinimumSize(QSize(500, 300))
        Dialog.setMaximumSize(QSize(500, 300))
        Dialog.setStyleSheet(u"background-image: url(:/splash/splash.png);")
        self.credits = QTextBrowser(Dialog)
        self.credits.setObjectName(u"credits")
        self.credits.setGeometry(QRect(20, 20, 211, 201))
        sizePolicy.setHeightForWidth(self.credits.sizePolicy().hasHeightForWidth())
        self.credits.setSizePolicy(sizePolicy)
        self.credits.setStyleSheet(u"background-image: url();\n"
"background-color: rgba(255, 255, 255, 20);\n"
"color: rgb(255, 255, 255);\n"
"border-radius: 8px;\n"
"padding: 10px;")
        self.credits.setFrameShape(QFrame.Shape.NoFrame)
        self.credits.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.credits.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.credits.setOpenExternalLinks(True)
        self.label_version = QLabel(Dialog)
        self.label_version.setObjectName(u"label_version")
        self.label_version.setGeometry(QRect(0, 0, 431, 151))
        self.label_version.setStyleSheet(u"background-image: url(); \n"
"color: rgb(0, 174, 238);\n"
"font: 24pt \"Kozuka Gothic Pro\";")
        self.label_version.setAlignment(Qt.AlignmentFlag.AlignBottom|Qt.AlignmentFlag.AlignRight|Qt.AlignmentFlag.AlignTrailing)

        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"USDB Syncer", None))
        self.credits.setMarkdown(QCoreApplication.translate("Dialog", u"\n"
"\n"
"\n"
"**Thank you** for using USDB Syncer!\n"
"\n"
"Keep your karaoke song collection up to\n"
"date at the click of a button.\n"
"\n"
"***\n"
"\n"
"Licensed under the \n"
"[MIT License](https://mit-license.org/).\n"
"Find the source code on \n"
"[GitHub](https://github.com/bohning/usdb_syncer/).\n"
"\n"
"\n"
"***\n"
"\n"
"\n"
"**More info**\n"
"[USDB Syncer Wiki](https://github.com/bohning/usdb_syncer/wiki)\n"
"\n"
"\n"
"\n"
"\\***\n"
"\n"
"**Support**\n"
"You can support the developers by buying them some \n"
"[vegan pizza!](https://www.buymeacoffee.com/usdbsyncer)\n"
"\n"
"\n"
"\\***\n"
"\n"
"**Main Programmers**\n"
"bohning\n"
"RumovZ\n"
"\n"
"**Contributing Programmers**\n"
"mjhalwa\n"
"g3n35i5\n"
"BWagener\n"
"\n"
"***\n"
"\n"
"**\n"
"Application Icon**\n"
"rawpixel.com on Freepik\n"
"Website: \n"
"[freepik.com](https://www.freepik.com/free-vector/pink-neon-cloud-icon-digital-networking-system_16406257.htm)\n"
"**Fugue Icons**\n"
"\u00a9 2021 Yusuke Kamiyamane\n"
"Website: \n"
"[p.yusukekamiyama"
                        "ne.com/](https://p.yusukekamiyamane.com/)\n"
"License: \n"
"[CC Attribution 3.0 Unported](http://creativecommons.org/licenses/by/3.0)\n"
"\n"
"\\***\n"
"\n"
"**Testers**\n"
"Rakuri\n"
"Hoanzl\n"
"Gr\u00fcneNeun\n"
"\n"
"\n"
"\n"
"\n"
"**Copyright \u00a9 2023**\n"
"\n"
"", None))
        self.credits.setHtml(QCoreApplication.translate("Dialog", u"<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><meta charset=\"utf-8\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"hr { height: 1px; border-width: 0; }\n"
"li.unchecked::marker { content: \"\\2610\"; }\n"
"li.checked::marker { content: \"\\2612\"; }\n"
"</style></head><body style=\" font-family:'Segoe UI'; font-size:9pt; font-weight:400; font-style:normal;\">\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /><br /><br /><span style=\" font-weight:700;\">Thank you</span> for using USDB Syncer!<br /><br />Keep your karaoke song collection up to date at the click of a button.<br /><br />***<br /><br />Licensed under the <a href=\"https://mit-license.org/\"><span style=\" text-decoration: underline; color:#00aeef;\">MIT License</span></a>.<br />Find the source code on <a href=\"https://"
                        "github.com/bohning/usdb_syncer/\"><span style=\" text-decoration: underline; color:#00aeef;\">GitHub</span></a>.<br /><br /><br />***</p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /><span style=\" font-weight:700;\">More info</span><br /><a href=\"https://github.com/bohning/usdb_syncer/wiki\"><span style=\" text-decoration: underline; color:#00aeef;\">USDB Syncer Wiki</span></a><br /></p>\n"
"<p align=\"center\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br />***<br /><br /><span style=\" font-weight:700;\">Support</span><br />You can support the developers by buying them some <a href=\"https://www.buymeacoffee.com/usdbsyncer\"><span style=\" text-decoration: underline; color:#00aeef;\">vegan pizza!</span></a><br /><br /><br />***<br /><br /><span style=\" font-weight:700;\">Main Programmers</span><br />bohning<br />RumovZ<br /><br /"
                        "><span style=\" font-weight:700;\">Contributing Programmers</span><br />mjhalwa<br />g3n35i5<br />BWagener<br /><br />***<br /><br /><span style=\" font-weight:700;\">Application Icon</span><br />rawpixel.com on Freepik<br />Website: <a href=\"https://www.freepik.com/free-vector/pink-neon-cloud-icon-digital-networking-system_16406257.htm\"><span style=\" text-decoration: underline; color:#00aeef;\">freepik.com</span></a><br /><br /><span style=\" font-weight:700;\">Fugue Icons</span><br />\u00a9 2021 Yusuke Kamiyamane<br />Website: <a href=\"https://p.yusukekamiyamane.com/\"><span style=\" text-decoration: underline; color:#00aeef;\">p.yusukekamiyamane.com/</span></a><br />License: <a href=\"http://creativecommons.org/licenses/by/3.0\"><span style=\" text-decoration: underline; color:#00aeef;\">CC Attribution 3.0 Unported</span></a><br /><br />***<br /><br /><span style=\" font-weight:700;\">Testers</span><br />Rakuri<br />Hoanzl<br />Gr\u00fcneNeun<br /><br /><br /><br /><br /><span style=\" font-weight:700;\""
                        ">Copyright \u00a9 2023</span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; font-family:'.AppleSystemUIFont'; font-size:13pt;\"><br /></p></body></html>", None))
        self.label_version.setText(QCoreApplication.translate("Dialog", u"VERSION", None))
    # retranslateUi




"""Dialog to create a report."""

import datetime
from collections.abc import Iterable
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QFileDialog,
    QListWidgetItem,
    QMessageBox,
    QWidget,
)

from usdb_syncer import SongId, db, settings, utils
from usdb_syncer.gui import progress
from usdb_syncer.gui.forms.ReportDialog import Ui_Dialog
from usdb_syncer.gui.progress import run_with_progress
from usdb_syncer.gui.song_table.column import Column
from usdb_syncer.gui.song_table.song_table import SongTable
from usdb_syncer.json_export import generate_report_json
from usdb_syncer.logger import logger
from usdb_syncer.pdf import generate_report_pdf

optional_columns = [
    Column.LANGUAGE,
    Column.EDITION,
    Column.GENRE,
    Column.YEAR,
    Column.CREATOR,
    Column.SONG_ID,
]


class ReportDialog(Ui_Dialog, QDialog):
    """Dialog to create a report."""

    def __init__(self, parent: QWidget, song_table: SongTable) -> None:
        super().__init__(parent=parent)
        self.song_table = song_table
        self.setupUi(self)
        self._populate_comboboxes()
        self._populate_optional_columns()
        self._load_settings()

    def _populate_optional_columns(self) -> None:
        for column in optional_columns:
            item = QListWidgetItem()
            item.setData(Qt.ItemDataRole.UserRole, column)
            if icon := column.decoration_data():
                item.setIcon(icon)
            if name := column.display_data():
                item.setText(name)
            item.setCheckState(Qt.CheckState.Unchecked)
            if column in [Column.ARTIST, Column.TITLE]:
                item.setCheckState(Qt.CheckState.Checked)
            item.setFlags(
                Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsUserCheckable
                | Qt.ItemFlag.ItemIsSelectable
                | Qt.ItemFlag.ItemIsDragEnabled
            )
            self.optional_columns.addItem(item)
        self.optional_columns.setDragEnabled(True)
        self.optional_columns.setDragDropMode(
            QAbstractItemView.DragDropMode.InternalMove
        )

    def _populate_comboboxes(self) -> None:
        combobox_settings = (
            (self.comboBox_pdf_pagesize, settings.ReportPDFPagesize),
            (self.comboBox_pdf_orientation, settings.ReportPDFOrientation),
        )
        for combobox, setting in combobox_settings:
            for item in setting:
                combobox.addItem(str(item), item)

    def _load_settings(self) -> None:
        self.comboBox_pdf_pagesize.setCurrentIndex(
            self.comboBox_pdf_pagesize.findData(settings.get_report_pdf_pagesize())
        )
        self.comboBox_pdf_orientation.setCurrentIndex(
            self.comboBox_pdf_orientation.findData(
                settings.get_report_pdf_orientation()
            )
        )
        self.spinBox_pdf_margin.setValue(settings.get_report_pdf_margin())
        self.spinBox_pdf_columns.setValue(settings.get_report_pdf_columns())
        self.spinBox_pdf_font_size.setValue(settings.get_report_pdf_fontsize())
        self.spinBox_json_indent.setValue(settings.get_report_json_indent())

    def accept(self) -> None:
        self._save_settings()
        if self.tabWidget_report_type.currentIndex() == 0:
            if self._generate_report_pdf():
                super().accept()
        elif self.tabWidget_report_type.currentIndex() == 1:
            if self._generate_report_json():
                super().accept()

    def reject(self) -> None:
        self._save_settings()
        super().reject()

    def _save_settings(self) -> None:
        settings.set_report_pdf_pagesize(self.comboBox_pdf_pagesize.currentData())
        settings.set_report_pdf_orientation(self.comboBox_pdf_orientation.currentData())
        settings.set_report_pdf_margin(self.spinBox_pdf_margin.value())
        settings.set_report_pdf_columns(self.spinBox_pdf_columns.value())
        settings.set_report_pdf_fontsize(self.spinBox_pdf_font_size.value())
        settings.set_report_json_indent(self.spinBox_json_indent.value())

    def _generate_report_pdf(self) -> bool:
        def on_done(result: progress.Result) -> None:
            path = result.result()
            logger.info(f"PDF report created at {path}.")
            utils.open_path_or_file(path)

        songs: Iterable[SongId] = []
        if self.radioButton_locally_available_songs.isChecked():
            songs = db.all_local_usdb_songs()
        if self.radioButton_selected_songs.isChecked():
            songs = []
            for song in self.song_table.selected_songs():
                songs.append(song.song_id)
        elif self.radioButton_all_songs.isChecked():
            songs = db.all_song_ids()
        if not songs:
            msg = "Skipping PDF report creation: no songs match the selection."
            QMessageBox.information(self, "Empty song list", msg)
            logger.info(msg)
            return False
        fname = f"{datetime.datetime.now():%Y-%m-%d}_songlist.pdf"
        path = str(Path(settings.get_song_dir()) / fname)
        path = QFileDialog.getSaveFileName(self, dir=path, filter="PDF (*.pdf)")[0]
        if path:
            pagesize = self.comboBox_pdf_pagesize.currentData()
            orientation = self.comboBox_pdf_orientation.currentData()
            margin = self.spinBox_pdf_margin.value()
            column_count = self.spinBox_pdf_columns.value()
            base_font_size = self.spinBox_pdf_font_size.value()
            optional_info = []
            for index in range(self.optional_columns.count()):
                item: QListWidgetItem = self.optional_columns.item(index)
                if item.checkState() == Qt.CheckState.Checked:
                    optional_info.append(item.data(Qt.ItemDataRole.UserRole))
            run_with_progress(
                "Creating PDF report ...",
                lambda: generate_report_pdf(
                    songs=songs,
                    path=path,
                    size=pagesize,
                    orientation=orientation,
                    margin=margin,
                    column_count=column_count,
                    base_font_size=base_font_size,
                    optional_info=optional_info,
                ),
                on_done=on_done,
            )
            return True
        # dialog is hidden by main window on macOS if file picker was cancelled
        self.raise_()
        return False

    def _generate_report_json(self) -> bool:
        def on_done(result: progress.Result) -> None:
            path, num_of_songs = result.result()
            logger.info(f"JSON report created at {path} ({num_of_songs} songs).")

        songs: Iterable[SongId] = []
        if self.radioButton_locally_available_songs.isChecked():
            songs = db.all_local_usdb_songs()
        if self.radioButton_selected_songs.isChecked():
            songs = []
            for song in self.song_table.selected_songs():
                songs.append(song.song_id)
        elif self.radioButton_all_songs.isChecked():
            songs = db.all_song_ids()
        songs = db.all_local_usdb_songs()
        if not songs:
            msg = "Skipping JSON report creation: no songs match the selection."
            QMessageBox.information(self, "Empty song list", msg)
            logger.info(msg)
            return False
        fname = f"{datetime.datetime.now():%Y-%m-%d}_songlist.json"
        path = str(Path(settings.get_song_dir()) / fname)
        path = QFileDialog.getSaveFileName(self, dir=path, filter="JSON (*.json)")[0]
        indent = self.spinBox_json_indent.value()
        if path:
            run_with_progress(
                "Creating JSON report ...",
                lambda: generate_report_json(
                    songs=songs, path=Path(path), indent=indent
                ),
                on_done=on_done,
            )
            return True
        # dialog is hidden by main window on macOS if file picker was cancelled
        self.raise_()
        return False



"""Utilities for managing dynamically loaded add-ons."""

import importlib
import pkgutil
import sys

from usdb_syncer import utils
from usdb_syncer.logger import logger


def load_all() -> None:
    addons_path = utils.AppPaths.addons
    sys.path.append(str(addons_path))

    # Search for zip files in the addons directory and add them to the path
    zip_files = [str(z) for z in addons_path.glob("*.zip")]
    for zip_file in zip_files:
        sys.path.append(zip_file)

    # Import everything
    for _, mod, is_pkg in pkgutil.iter_modules([*zip_files, str(addons_path)]):
        importlib.import_module(mod)
        logger.debug(f"Imported add-on {'package' if is_pkg else 'module'} '{mod}'.")



"""Generates a PDF from the passed song list."""

from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from importlib import resources
from typing import assert_never

from reportlab.lib import colors
from reportlab.lib.pagesizes import A3, A4, A5, LEGAL, LETTER, landscape
from reportlab.lib.styles import ParagraphStyle
from reportlab.lib.units import mm
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from reportlab.pdfgen.canvas import Canvas
from reportlab.platypus import BaseDocTemplate, Flowable, Frame, PageTemplate, Paragraph

from usdb_syncer import SongId, utils
from usdb_syncer.gui.resources import fonts
from usdb_syncer.gui.song_table.column import Column
from usdb_syncer.settings import ReportPDFOrientation, ReportPDFPagesize
from usdb_syncer.usdb_song import UsdbSong

NOTOSANS_BLACK = "NotoSans-Black"
NOTOSANS_BOLD = "NotoSans-Bold"
NOTOSANS_REGULAR = "NotoSans-Regular"


_font_files = resources.files(fonts)
for name in (NOTOSANS_BLACK, NOTOSANS_BOLD, NOTOSANS_REGULAR):
    pdfmetrics.registerFont(TTFont(name, _font_files.joinpath(f"{name}.ttf")))


class Bookmark(Flowable):
    """Bookmark flowable."""

    def __init__(self, key: str, title: str, level: int = 0):
        super().__init__()
        self.key = key
        self.title = title
        self.level = level

    def draw(self) -> None:
        self.canv.bookmarkPage(self.key)
        self.canv.addOutlineEntry(self.title, self.key, level=self.level, closed=False)


def generate_report_pdf(
    *,
    songs: Iterable[SongId],
    path: str,
    size: ReportPDFPagesize = ReportPDFPagesize.A4,
    orientation: ReportPDFOrientation = ReportPDFOrientation.PORTRAIT,
    margin: int = 20,
    column_count: int = 1,
    base_font_size: int = 10,
    optional_info: list[Column] | None = None,
) -> str:
    optional_info = optional_info or []
    pagesize = _get_pagesize(size, orientation)

    doc: BaseDocTemplate = BaseDocTemplate(
        path,
        pagesize=pagesize,
        leftMargin=margin * mm,
        rightMargin=margin * mm,
        topMargin=margin * mm,
        bottomMargin=margin * mm,
    )

    custom_styles = _create_paragraph_styles(base_font_size)
    content = _build_pdf_content(songs, base_font_size, custom_styles, optional_info)

    gutter = margin * mm * 0.5
    frames = _create_column_frames(
        pagesize, margin=margin * mm, gutter=gutter, column_count=column_count
    )
    template = PageTemplate(
        id="with_page_number", frames=frames, onPage=_add_page_number
    )
    doc.addPageTemplates([template])
    doc.build(content)

    return path


def _get_pagesize(
    size: ReportPDFPagesize, orientation: ReportPDFOrientation
) -> tuple[float, float]:
    match size:
        case ReportPDFPagesize.A3:
            pagesize = A3
        case ReportPDFPagesize.A4:
            pagesize = A4
        case ReportPDFPagesize.A5:
            pagesize = A5
        case ReportPDFPagesize.LEGAL:
            pagesize = LEGAL
        case ReportPDFPagesize.LETTER:
            pagesize = LETTER
        case _ as unreachable:
            assert_never(unreachable)
    return (
        landscape(pagesize)
        if orientation == ReportPDFOrientation.LANDSCAPE
        else pagesize
    )


@dataclass(frozen=True)
class ParagraphStyles:
    """Paragraph styles for the report."""

    initial: ParagraphStyle
    artist: ParagraphStyle
    entry: ParagraphStyle


def _create_paragraph_styles(base_font_size: int) -> ParagraphStyles:
    return ParagraphStyles(
        initial=ParagraphStyle(
            "Initial",
            fontName=NOTOSANS_BLACK,
            fontSize=base_font_size * 3,
            textColor=colors.green,
            spaceBefore=base_font_size * 2.4,
            spaceAfter=base_font_size * 2.4,
        ),
        artist=ParagraphStyle(
            "Artist",
            fontName=NOTOSANS_BOLD,
            fontSize=base_font_size * 1.2,
            spaceBefore=base_font_size * 1.2,
            leading=base_font_size * 1.6,
            textColor=colors.black,
            leftIndent=0,
        ),
        entry=ParagraphStyle(
            "Entry",
            fontName=NOTOSANS_REGULAR,
            fontSize=base_font_size,
            leftIndent=base_font_size,
            leading=base_font_size * 1.4,
        ),
    )


def _build_pdf_content(
    songs: Iterable[SongId],
    base_font_size: int,
    styles: ParagraphStyles,
    optional_info: list[Column],
) -> list:
    content: list = []
    artist_map: dict[str, list[UsdbSong]] = defaultdict(list)
    for song_id in songs:
        song = UsdbSong.get(song_id)
        if song:
            artist_map[song.artist].append(song)

    initial_map: dict[str, dict[str, list[UsdbSong]]] = defaultdict(dict)
    for artist, artist_songs in artist_map.items():
        initial = utils.get_first_alphanum_upper(artist) or artist[0].upper()
        initial_map[initial][artist] = sorted(
            artist_songs, key=lambda s: s.title.lower()
        )

    for initial in sorted(initial_map.keys()):
        bookmark_key = f"initial_{initial}"
        content.append(Bookmark(bookmark_key, initial))
        content.append(Paragraph(initial, styles.initial))
        for artist in sorted(initial_map[initial].keys(), key=str.lower):
            content.append(Paragraph(f"<b>{artist}</b>", styles.artist))
            for song in initial_map[initial][artist]:
                entry = _format_song_entry(song, base_font_size, optional_info)
                content.append(Paragraph(entry, style=styles.entry))

    return content


def _format_song_entry(  # noqa: C901
    song: UsdbSong, base_font_size: int, optional_info: list[Column]
) -> str:
    entry = f"{song.title} "
    entry += f'<font size="{base_font_size * 0.8}" color="{colors.grey.hexval()}">'
    for column in optional_info:
        match column:
            case Column.LANGUAGE:
                if song.language:
                    entry += f"&nbsp;&nbsp;{song.language}"
            case Column.EDITION:
                if song.edition:
                    entry += f"&nbsp;&nbsp;{song.edition}"
            case Column.GENRE:
                if song.genre:
                    entry += f"&nbsp;&nbsp;{song.genre}"
            case Column.YEAR:
                if song.year:
                    entry += f"&nbsp;&nbsp;{song.year}"
            case Column.CREATOR:
                if song.creator:
                    entry += f"&nbsp;&nbsp;{song.creator}"
            case Column.SONG_ID:
                entry += f"&nbsp;&nbsp;({song.song_id})"
            case _:
                pass
    entry += "</font>"
    return entry


def _add_page_number(canvas: Canvas, doc: BaseDocTemplate) -> None:
    canvas.saveState()
    page_num: str = str(doc.page)
    canvas.setFont(NOTOSANS_REGULAR, 8)
    canvas.setFillColor(colors.grey)
    canvas.drawCentredString(doc.pagesize[0] / 2, doc.bottomMargin * 0.5, page_num)
    canvas.restoreState()


def _create_column_frames(
    pagesize: tuple[float, float], margin: float, gutter: float, column_count: int
) -> list[Frame]:
    usable_width = pagesize[0] - 2 * margin - (column_count - 1) * gutter
    column_width = usable_width / column_count
    frames: list[Frame] = []

    for i in range(column_count):
        x = margin + i * (column_width + gutter)
        frame = Frame(
            x, margin, column_width, pagesize[1] - 2 * margin, id=f"col{i + 1}"
        )
        frames.append(frame)

    return frames



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'MetaTagsDialog.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QDialog,
    QDialogButtonBox, QDoubleSpinBox, QFrame, QGroupBox,
    QHBoxLayout, QLabel, QLayout, QLineEdit,
    QPushButton, QScrollArea, QSizePolicy, QSpacerItem,
    QSpinBox, QVBoxLayout, QWidget)

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(861, 868)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setMinimumSize(QSize(0, 0))
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.scrollArea = QScrollArea(Dialog)
        self.scrollArea.setObjectName(u"scrollArea")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.scrollArea.sizePolicy().hasHeightForWidth())
        self.scrollArea.setSizePolicy(sizePolicy1)
        self.scrollArea.setFrameShape(QFrame.Shape.NoFrame)
        self.scrollArea.setFrameShadow(QFrame.Shadow.Plain)
        self.scrollArea.setLineWidth(0)
        self.scrollArea.setWidgetResizable(True)
        self.scrollAreaWidgetContents = QWidget()
        self.scrollAreaWidgetContents.setObjectName(u"scrollAreaWidgetContents")
        self.scrollAreaWidgetContents.setGeometry(QRect(0, 0, 843, 850))
        sizePolicy1.setHeightForWidth(self.scrollAreaWidgetContents.sizePolicy().hasHeightForWidth())
        self.scrollAreaWidgetContents.setSizePolicy(sizePolicy1)
        self.verticalLayout_2 = QVBoxLayout(self.scrollAreaWidgetContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setSizeConstraint(QLayout.SizeConstraint.SetNoConstraint)
        self.audio = QGroupBox(self.scrollAreaWidgetContents)
        self.audio.setObjectName(u"audio")
        self.verticalLayout_5 = QVBoxLayout(self.audio)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.verticalLayout_5.setContentsMargins(4, 4, 4, 4)
        self.audio_layout = QHBoxLayout()
        self.audio_layout.setObjectName(u"audio_layout")
        self.audio_label = QLabel(self.audio)
        self.audio_label.setObjectName(u"audio_label")

        self.audio_layout.addWidget(self.audio_label)

        self.audio_url = QLineEdit(self.audio)
        self.audio_url.setObjectName(u"audio_url")

        self.audio_layout.addWidget(self.audio_url)


        self.verticalLayout_5.addLayout(self.audio_layout)


        self.verticalLayout_2.addWidget(self.audio)

        self.video = QGroupBox(self.scrollAreaWidgetContents)
        self.video.setObjectName(u"video")
        self.verticalLayout_4 = QVBoxLayout(self.video)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.verticalLayout_4.setContentsMargins(4, 4, 4, 4)
        self.video_layout = QHBoxLayout()
        self.video_layout.setObjectName(u"video_layout")
        self.video_label = QLabel(self.video)
        self.video_label.setObjectName(u"video_label")

        self.video_layout.addWidget(self.video_label)

        self.video_url = QLineEdit(self.video)
        self.video_url.setObjectName(u"video_url")

        self.video_layout.addWidget(self.video_url)


        self.verticalLayout_4.addLayout(self.video_layout)


        self.verticalLayout_2.addWidget(self.video)

        self.cover = QGroupBox(self.scrollAreaWidgetContents)
        self.cover.setObjectName(u"cover")
        self.verticalLayout_6 = QVBoxLayout(self.cover)
        self.verticalLayout_6.setSpacing(6)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.verticalLayout_6.setContentsMargins(4, 4, 4, 4)
        self.cover_url_layout = QHBoxLayout()
        self.cover_url_layout.setObjectName(u"cover_url_layout")
        self.cover_label = QLabel(self.cover)
        self.cover_label.setObjectName(u"cover_label")

        self.cover_url_layout.addWidget(self.cover_label)

        self.cover_url = QLineEdit(self.cover)
        self.cover_url.setObjectName(u"cover_url")

        self.cover_url_layout.addWidget(self.cover_url)


        self.verticalLayout_6.addLayout(self.cover_url_layout)

        self.cover_rotate_contrast_resize_layout = QHBoxLayout()
        self.cover_rotate_contrast_resize_layout.setObjectName(u"cover_rotate_contrast_resize_layout")
        self.cover_rotate_contrast_resize_layout.setContentsMargins(-1, -1, -1, 0)
        self.groupBox_12 = QGroupBox(self.cover)
        self.groupBox_12.setObjectName(u"groupBox_12")
        self.horizontalLayout_15 = QHBoxLayout(self.groupBox_12)
        self.horizontalLayout_15.setObjectName(u"horizontalLayout_15")
        self.horizontalLayout_15.setContentsMargins(4, 4, 4, 4)
        self.rotation_label = QLabel(self.groupBox_12)
        self.rotation_label.setObjectName(u"rotation_label")

        self.horizontalLayout_15.addWidget(self.rotation_label)

        self.cover_rotation = QDoubleSpinBox(self.groupBox_12)
        self.cover_rotation.setObjectName(u"cover_rotation")
        self.cover_rotation.setMinimum(-360.000000000000000)
        self.cover_rotation.setMaximum(360.000000000000000)
        self.cover_rotation.setSingleStep(0.100000000000000)

        self.horizontalLayout_15.addWidget(self.cover_rotation)

        self.horizontalSpacer_10 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_15.addItem(self.horizontalSpacer_10)


        self.cover_rotate_contrast_resize_layout.addWidget(self.groupBox_12)

        self.groupBox_14 = QGroupBox(self.cover)
        self.groupBox_14.setObjectName(u"groupBox_14")
        self.horizontalLayout_19 = QHBoxLayout(self.groupBox_14)
        self.horizontalLayout_19.setObjectName(u"horizontalLayout_19")
        self.horizontalLayout_19.setContentsMargins(4, 4, 4, 4)
        self.contrast_label = QLabel(self.groupBox_14)
        self.contrast_label.setObjectName(u"contrast_label")

        self.horizontalLayout_19.addWidget(self.contrast_label)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.cover_contrast_auto = QCheckBox(self.groupBox_14)
        self.cover_contrast_auto.setObjectName(u"cover_contrast_auto")

        self.horizontalLayout_3.addWidget(self.cover_contrast_auto)

        self.cover_contrast = QDoubleSpinBox(self.groupBox_14)
        self.cover_contrast.setObjectName(u"cover_contrast")
        self.cover_contrast.setSingleStep(0.100000000000000)
        self.cover_contrast.setValue(1.000000000000000)

        self.horizontalLayout_3.addWidget(self.cover_contrast)


        self.horizontalLayout_19.addLayout(self.horizontalLayout_3)

        self.horizontalSpacer_12 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_19.addItem(self.horizontalSpacer_12)


        self.cover_rotate_contrast_resize_layout.addWidget(self.groupBox_14)

        self.groupBox_13 = QGroupBox(self.cover)
        self.groupBox_13.setObjectName(u"groupBox_13")
        self.horizontalLayout_16 = QHBoxLayout(self.groupBox_13)
        self.horizontalLayout_16.setObjectName(u"horizontalLayout_16")
        self.horizontalLayout_16.setContentsMargins(4, 4, 4, 4)
        self.resize_label = QLabel(self.groupBox_13)
        self.resize_label.setObjectName(u"resize_label")

        self.horizontalLayout_16.addWidget(self.resize_label)

        self.cover_resize = QSpinBox(self.groupBox_13)
        self.cover_resize.setObjectName(u"cover_resize")
        self.cover_resize.setMinimum(0)
        self.cover_resize.setMaximum(1920)
        self.cover_resize.setValue(0)

        self.horizontalLayout_16.addWidget(self.cover_resize)

        self.horizontalSpacer_11 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_16.addItem(self.horizontalSpacer_11)


        self.cover_rotate_contrast_resize_layout.addWidget(self.groupBox_13)


        self.verticalLayout_6.addLayout(self.cover_rotate_contrast_resize_layout)

        self.cover_crop_group = QGroupBox(self.cover)
        self.cover_crop_group.setObjectName(u"cover_crop_group")
        self.horizontalLayout_2 = QHBoxLayout(self.cover_crop_group)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(4, 4, 4, 4)
        self.cover_crop_left_label = QLabel(self.cover_crop_group)
        self.cover_crop_left_label.setObjectName(u"cover_crop_left_label")
        sizePolicy1.setHeightForWidth(self.cover_crop_left_label.sizePolicy().hasHeightForWidth())
        self.cover_crop_left_label.setSizePolicy(sizePolicy1)

        self.horizontalLayout_2.addWidget(self.cover_crop_left_label)

        self.cover_crop_left = QSpinBox(self.cover_crop_group)
        self.cover_crop_left.setObjectName(u"cover_crop_left")
        self.cover_crop_left.setMaximum(9999)

        self.horizontalLayout_2.addWidget(self.cover_crop_left)

        self.cover_crop_top_label = QLabel(self.cover_crop_group)
        self.cover_crop_top_label.setObjectName(u"cover_crop_top_label")

        self.horizontalLayout_2.addWidget(self.cover_crop_top_label)

        self.cover_crop_top = QSpinBox(self.cover_crop_group)
        self.cover_crop_top.setObjectName(u"cover_crop_top")
        self.cover_crop_top.setMaximum(9999)

        self.horizontalLayout_2.addWidget(self.cover_crop_top)

        self.cover_crop_width_label = QLabel(self.cover_crop_group)
        self.cover_crop_width_label.setObjectName(u"cover_crop_width_label")

        self.horizontalLayout_2.addWidget(self.cover_crop_width_label)

        self.cover_crop_width = QSpinBox(self.cover_crop_group)
        self.cover_crop_width.setObjectName(u"cover_crop_width")
        self.cover_crop_width.setMaximum(9999)

        self.horizontalLayout_2.addWidget(self.cover_crop_width)

        self.cover_crop_height_label = QLabel(self.cover_crop_group)
        self.cover_crop_height_label.setObjectName(u"cover_crop_height_label")

        self.horizontalLayout_2.addWidget(self.cover_crop_height_label)

        self.cover_crop_height = QSpinBox(self.cover_crop_group)
        self.cover_crop_height.setObjectName(u"cover_crop_height")
        self.cover_crop_height.setMaximum(9999)

        self.horizontalLayout_2.addWidget(self.cover_crop_height)

        self.horizontalSpacer_4 = QSpacerItem(0, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_4)


        self.verticalLayout_6.addWidget(self.cover_crop_group)


        self.verticalLayout_2.addWidget(self.cover)

        self.background = QGroupBox(self.scrollAreaWidgetContents)
        self.background.setObjectName(u"background")
        self.verticalLayout_7 = QVBoxLayout(self.background)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(4, 4, 4, 4)
        self.background_url_layout = QHBoxLayout()
        self.background_url_layout.setObjectName(u"background_url_layout")
        self.background_label = QLabel(self.background)
        self.background_label.setObjectName(u"background_label")

        self.background_url_layout.addWidget(self.background_label)

        self.background_url = QLineEdit(self.background)
        self.background_url.setObjectName(u"background_url")

        self.background_url_layout.addWidget(self.background_url)


        self.verticalLayout_7.addLayout(self.background_url_layout)

        self.background_resize_group = QGroupBox(self.background)
        self.background_resize_group.setObjectName(u"background_resize_group")
        self.horizontalLayout_9 = QHBoxLayout(self.background_resize_group)
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.horizontalLayout_9.setContentsMargins(4, 4, 4, 4)
        self.background_resize_width_label = QLabel(self.background_resize_group)
        self.background_resize_width_label.setObjectName(u"background_resize_width_label")

        self.horizontalLayout_9.addWidget(self.background_resize_width_label)

        self.background_resize_width = QSpinBox(self.background_resize_group)
        self.background_resize_width.setObjectName(u"background_resize_width")
        self.background_resize_width.setMaximum(9999)

        self.horizontalLayout_9.addWidget(self.background_resize_width)

        self.label_22background_resize_height_label = QLabel(self.background_resize_group)
        self.label_22background_resize_height_label.setObjectName(u"label_22background_resize_height_label")

        self.horizontalLayout_9.addWidget(self.label_22background_resize_height_label)

        self.background_resize_height = QSpinBox(self.background_resize_group)
        self.background_resize_height.setObjectName(u"background_resize_height")
        self.background_resize_height.setMaximum(9999)

        self.horizontalLayout_9.addWidget(self.background_resize_height)

        self.horizontalSpacer_9 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_9.addItem(self.horizontalSpacer_9)


        self.verticalLayout_7.addWidget(self.background_resize_group)

        self.background_crop_group = QGroupBox(self.background)
        self.background_crop_group.setObjectName(u"background_crop_group")
        self.horizontalLayout_4 = QHBoxLayout(self.background_crop_group)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(4, 4, 4, 4)
        self.background_crop_left_label = QLabel(self.background_crop_group)
        self.background_crop_left_label.setObjectName(u"background_crop_left_label")
        sizePolicy1.setHeightForWidth(self.background_crop_left_label.sizePolicy().hasHeightForWidth())
        self.background_crop_left_label.setSizePolicy(sizePolicy1)

        self.horizontalLayout_4.addWidget(self.background_crop_left_label)

        self.background_crop_left = QSpinBox(self.background_crop_group)
        self.background_crop_left.setObjectName(u"background_crop_left")
        self.background_crop_left.setMaximum(9999)

        self.horizontalLayout_4.addWidget(self.background_crop_left)

        self.background_crop_top_label = QLabel(self.background_crop_group)
        self.background_crop_top_label.setObjectName(u"background_crop_top_label")

        self.horizontalLayout_4.addWidget(self.background_crop_top_label)

        self.background_crop_top = QSpinBox(self.background_crop_group)
        self.background_crop_top.setObjectName(u"background_crop_top")
        self.background_crop_top.setMaximum(9999)

        self.horizontalLayout_4.addWidget(self.background_crop_top)

        self.background_crop_width_label = QLabel(self.background_crop_group)
        self.background_crop_width_label.setObjectName(u"background_crop_width_label")

        self.horizontalLayout_4.addWidget(self.background_crop_width_label)

        self.background_crop_width = QSpinBox(self.background_crop_group)
        self.background_crop_width.setObjectName(u"background_crop_width")
        self.background_crop_width.setMaximum(9999)

        self.horizontalLayout_4.addWidget(self.background_crop_width)

        self.background_crop_height_label = QLabel(self.background_crop_group)
        self.background_crop_height_label.setObjectName(u"background_crop_height_label")

        self.horizontalLayout_4.addWidget(self.background_crop_height_label)

        self.background_crop_height = QSpinBox(self.background_crop_group)
        self.background_crop_height.setObjectName(u"background_crop_height")
        self.background_crop_height.setMaximum(9999)

        self.horizontalLayout_4.addWidget(self.background_crop_height)

        self.horizontalSpacer_8 = QSpacerItem(385, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer_8)


        self.verticalLayout_7.addWidget(self.background_crop_group)


        self.verticalLayout_2.addWidget(self.background)

        self.duet = QGroupBox(self.scrollAreaWidgetContents)
        self.duet.setObjectName(u"duet")
        sizePolicy1.setHeightForWidth(self.duet.sizePolicy().hasHeightForWidth())
        self.duet.setSizePolicy(sizePolicy1)
        self.duet.setCheckable(True)
        self.duet.setChecked(False)
        self.horizontalLayout_6 = QHBoxLayout(self.duet)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.horizontalLayout_6.setContentsMargins(4, 4, 4, 4)
        self.duet_p1_label = QLabel(self.duet)
        self.duet_p1_label.setObjectName(u"duet_p1_label")
        sizePolicy1.setHeightForWidth(self.duet_p1_label.sizePolicy().hasHeightForWidth())
        self.duet_p1_label.setSizePolicy(sizePolicy1)
        self.duet_p1_label.setMaximumSize(QSize(16777215, 26))

        self.horizontalLayout_6.addWidget(self.duet_p1_label)

        self.duet_p1 = QLineEdit(self.duet)
        self.duet_p1.setObjectName(u"duet_p1")

        self.horizontalLayout_6.addWidget(self.duet_p1)

        self.duet_p2_label = QLabel(self.duet)
        self.duet_p2_label.setObjectName(u"duet_p2_label")
        sizePolicy1.setHeightForWidth(self.duet_p2_label.sizePolicy().hasHeightForWidth())
        self.duet_p2_label.setSizePolicy(sizePolicy1)
        self.duet_p2_label.setMaximumSize(QSize(16777215, 26))

        self.horizontalLayout_6.addWidget(self.duet_p2_label)

        self.duet_p2 = QLineEdit(self.duet)
        self.duet_p2.setObjectName(u"duet_p2")

        self.horizontalLayout_6.addWidget(self.duet_p2)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_3)


        self.verticalLayout_2.addWidget(self.duet)

        self.preview_medley = QHBoxLayout()
        self.preview_medley.setObjectName(u"preview_medley")
        self.preview = QGroupBox(self.scrollAreaWidgetContents)
        self.preview.setObjectName(u"preview")
        self.horizontalLayout_7 = QHBoxLayout(self.preview)
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.preview_start_label = QLabel(self.preview)
        self.preview_start_label.setObjectName(u"preview_start_label")

        self.horizontalLayout_7.addWidget(self.preview_start_label)

        self.preview_start = QDoubleSpinBox(self.preview)
        self.preview_start.setObjectName(u"preview_start")
        self.preview_start.setDecimals(3)
        self.preview_start.setMaximum(999.990000000000009)

        self.horizontalLayout_7.addWidget(self.preview_start)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer_2)


        self.preview_medley.addWidget(self.preview)

        self.medley = QGroupBox(self.scrollAreaWidgetContents)
        self.medley.setObjectName(u"medley")
        self.horizontalLayout_11 = QHBoxLayout(self.medley)
        self.horizontalLayout_11.setObjectName(u"horizontalLayout_11")
        self.medley_start_label = QLabel(self.medley)
        self.medley_start_label.setObjectName(u"medley_start_label")

        self.horizontalLayout_11.addWidget(self.medley_start_label)

        self.medley_start = QSpinBox(self.medley)
        self.medley_start.setObjectName(u"medley_start")
        self.medley_start.setMaximum(99999)

        self.horizontalLayout_11.addWidget(self.medley_start)

        self.medley_end_label = QLabel(self.medley)
        self.medley_end_label.setObjectName(u"medley_end_label")

        self.horizontalLayout_11.addWidget(self.medley_end_label)

        self.medley_end = QSpinBox(self.medley)
        self.medley_end.setObjectName(u"medley_end")
        self.medley_end.setMaximum(99999)

        self.horizontalLayout_11.addWidget(self.medley_end)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_11.addItem(self.horizontalSpacer_5)


        self.preview_medley.addWidget(self.medley)


        self.verticalLayout_2.addLayout(self.preview_medley)

        self.tags_2 = QGroupBox(self.scrollAreaWidgetContents)
        self.tags_2.setObjectName(u"tags_2")
        self.verticalLayout_3 = QVBoxLayout(self.tags_2)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(4, 4, 4, 4)
        self.tags_layout = QHBoxLayout()
        self.tags_layout.setObjectName(u"tags_layout")
        self.tags_label = QLabel(self.tags_2)
        self.tags_label.setObjectName(u"tags_label")

        self.tags_layout.addWidget(self.tags_label)

        self.tags = QLineEdit(self.tags_2)
        self.tags.setObjectName(u"tags")

        self.tags_layout.addWidget(self.tags)


        self.verticalLayout_3.addLayout(self.tags_layout)


        self.verticalLayout_2.addWidget(self.tags_2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)

        self.meta_tags = QHBoxLayout()
        self.meta_tags.setObjectName(u"meta_tags")
        self.output = QLineEdit(self.scrollAreaWidgetContents)
        self.output.setObjectName(u"output")
        self.output.setReadOnly(True)

        self.meta_tags.addWidget(self.output)

        self.button_copy_to_clipboard = QPushButton(self.scrollAreaWidgetContents)
        self.button_copy_to_clipboard.setObjectName(u"button_copy_to_clipboard")

        self.meta_tags.addWidget(self.button_copy_to_clipboard)


        self.verticalLayout_2.addLayout(self.meta_tags)

        self.buttonBox = QDialogButtonBox(self.scrollAreaWidgetContents)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)

        self.verticalLayout_2.addWidget(self.buttonBox)

        self.scrollArea.setWidget(self.scrollAreaWidgetContents)

        self.verticalLayout.addWidget(self.scrollArea)


        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Meta Tags", None))
        self.audio.setTitle(QCoreApplication.translate("Dialog", u"Audio (a)", None))
        self.audio_label.setText(QCoreApplication.translate("Dialog", u"URL:", None))
        self.video.setTitle(QCoreApplication.translate("Dialog", u"Video (v)", None))
        self.video_label.setText(QCoreApplication.translate("Dialog", u"URL:", None))
        self.cover.setTitle(QCoreApplication.translate("Dialog", u"Cover (co)", None))
        self.cover_label.setText(QCoreApplication.translate("Dialog", u"URL:", None))
        self.groupBox_12.setTitle(QCoreApplication.translate("Dialog", u"Rotation (co-rotate)", None))
        self.rotation_label.setText(QCoreApplication.translate("Dialog", u"Angle (ccw):", None))
        self.cover_rotation.setSuffix(QCoreApplication.translate("Dialog", u"\u00b0", None))
        self.groupBox_14.setTitle(QCoreApplication.translate("Dialog", u"Contrast (co-contrast)", None))
        self.contrast_label.setText(QCoreApplication.translate("Dialog", u"Value:", None))
        self.cover_contrast_auto.setText(QCoreApplication.translate("Dialog", u"Auto", None))
        self.groupBox_13.setTitle(QCoreApplication.translate("Dialog", u"Resize (co-resize)", None))
        self.resize_label.setText(QCoreApplication.translate("Dialog", u"Width/Height:", None))
        self.cover_resize.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.cover_crop_group.setTitle(QCoreApplication.translate("Dialog", u"Crop (co-crop)", None))
        self.cover_crop_left_label.setText(QCoreApplication.translate("Dialog", u"Left:", None))
        self.cover_crop_left.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.cover_crop_top_label.setText(QCoreApplication.translate("Dialog", u"Top:", None))
        self.cover_crop_top.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.cover_crop_width_label.setText(QCoreApplication.translate("Dialog", u"Width:", None))
        self.cover_crop_width.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.cover_crop_height_label.setText(QCoreApplication.translate("Dialog", u"Height:", None))
        self.cover_crop_height.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.background.setTitle(QCoreApplication.translate("Dialog", u"Background (bg)", None))
        self.background_label.setText(QCoreApplication.translate("Dialog", u"URL:", None))
        self.background_resize_group.setTitle(QCoreApplication.translate("Dialog", u"Resize (bg-resize)", None))
        self.background_resize_width_label.setText(QCoreApplication.translate("Dialog", u"Width:", None))
        self.background_resize_width.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.label_22background_resize_height_label.setText(QCoreApplication.translate("Dialog", u"Height:", None))
        self.background_resize_height.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.background_crop_group.setTitle(QCoreApplication.translate("Dialog", u"Crop (bg-crop)", None))
        self.background_crop_left_label.setText(QCoreApplication.translate("Dialog", u"Left:", None))
        self.background_crop_left.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.background_crop_top_label.setText(QCoreApplication.translate("Dialog", u"Top:", None))
        self.background_crop_top.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.background_crop_width_label.setText(QCoreApplication.translate("Dialog", u"Width:", None))
        self.background_crop_width.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.background_crop_height_label.setText(QCoreApplication.translate("Dialog", u"Height:", None))
        self.background_crop_height.setSuffix(QCoreApplication.translate("Dialog", u" px", None))
        self.duet.setTitle(QCoreApplication.translate("Dialog", u"Duet (p1, p2)", None))
        self.duet_p1_label.setText(QCoreApplication.translate("Dialog", u"Player 1:", None))
        self.duet_p1.setText(QCoreApplication.translate("Dialog", u"P1", None))
        self.duet_p2_label.setText(QCoreApplication.translate("Dialog", u"Player 2:", None))
        self.duet_p2.setText(QCoreApplication.translate("Dialog", u"P2", None))
        self.preview.setTitle(QCoreApplication.translate("Dialog", u"Preview (preview)", None))
        self.preview_start_label.setText(QCoreApplication.translate("Dialog", u"Preview start:", None))
        self.preview_start.setPrefix("")
        self.preview_start.setSuffix(QCoreApplication.translate("Dialog", u" s", None))
        self.medley.setTitle(QCoreApplication.translate("Dialog", u"Medley (medley)", None))
        self.medley_start_label.setText(QCoreApplication.translate("Dialog", u"Start:", None))
        self.medley_start.setSuffix("")
        self.medley_start.setPrefix(QCoreApplication.translate("Dialog", u"beat ", None))
        self.medley_end_label.setText(QCoreApplication.translate("Dialog", u"End:", None))
        self.medley_end.setPrefix(QCoreApplication.translate("Dialog", u"beat ", None))
        self.tags_2.setTitle(QCoreApplication.translate("Dialog", u"Tags (tags)", None))
        self.tags_label.setText(QCoreApplication.translate("Dialog", u"Tags:", None))
        self.output.setText(QCoreApplication.translate("Dialog", u"#VIDEO:", None))
        self.button_copy_to_clipboard.setText(QCoreApplication.translate("Dialog", u"Copy to Clipboard", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'MainWindow.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QDockWidget, QFrame,
    QHBoxLayout, QHeaderView, QLabel, QLineEdit,
    QMainWindow, QMenu, QMenuBar, QPlainTextEdit,
    QProgressBar, QPushButton, QSizePolicy, QSpacerItem,
    QStatusBar, QTableView, QToolButton, QTreeView,
    QVBoxLayout, QWidget)
from usdb_syncer.gui.resources.qt import resources as resources_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(1540, 838)
        self.action_meta_tags = QAction(MainWindow)
        self.action_meta_tags.setObjectName(u"action_meta_tags")
        icon = QIcon()
        icon.addFile(u":/icons/tag-hash.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_meta_tags.setIcon(icon)
        self.action_settings = QAction(MainWindow)
        self.action_settings.setObjectName(u"action_settings")
        icon1 = QIcon()
        icon1.addFile(u":/icons/cog.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_settings.setIcon(icon1)
        self.action_generate_song_list = QAction(MainWindow)
        self.action_generate_song_list.setObjectName(u"action_generate_song_list")
        icon2 = QIcon()
        icon2.addFile(u":/icons/report.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_generate_song_list.setIcon(icon2)
        self.action_refetch_song_list = QAction(MainWindow)
        self.action_refetch_song_list.setObjectName(u"action_refetch_song_list")
        icon3 = QIcon()
        icon3.addFile(u":/icons/check_for_update.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_refetch_song_list.setIcon(icon3)
        self.action_find_local_songs = QAction(MainWindow)
        self.action_find_local_songs.setObjectName(u"action_find_local_songs")
        icon4 = QIcon()
        icon4.addFile(u":/icons/database.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_find_local_songs.setIcon(icon4)
        self.action_show_log = QAction(MainWindow)
        self.action_show_log.setObjectName(u"action_show_log")
        icon5 = QIcon()
        icon5.addFile(u":/icons/log.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_show_log.setIcon(icon5)
        self.action_import_usdb_ids = QAction(MainWindow)
        self.action_import_usdb_ids.setObjectName(u"action_import_usdb_ids")
        icon6 = QIcon()
        icon6.addFile(u":/icons/document-export.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_import_usdb_ids.setIcon(icon6)
        self.action_export_usdb_ids = QAction(MainWindow)
        self.action_export_usdb_ids.setObjectName(u"action_export_usdb_ids")
        icon7 = QIcon()
        icon7.addFile(u":/icons/document-import.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_export_usdb_ids.setIcon(icon7)
        self.action_songs_download = QAction(MainWindow)
        self.action_songs_download.setObjectName(u"action_songs_download")
        icon8 = QIcon()
        icon8.addFile(u":/icons/drive-download.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_songs_download.setIcon(icon8)
        self.action_usdb_login = QAction(MainWindow)
        self.action_usdb_login.setObjectName(u"action_usdb_login")
        icon9 = QIcon()
        icon9.addFile(u":/icons/faviconUSDB.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_usdb_login.setIcon(icon9)
        self.action_about = QAction(MainWindow)
        self.action_about.setObjectName(u"action_about")
        icon10 = QIcon()
        icon10.addFile(u":/app/appicon_128x128.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_about.setIcon(icon10)
        self.action_show_in_usdb = QAction(MainWindow)
        self.action_show_in_usdb.setObjectName(u"action_show_in_usdb")
        self.action_show_in_usdb.setIcon(icon9)
        self.action_post_comment_in_usdb = QAction(MainWindow)
        self.action_post_comment_in_usdb.setObjectName(u"action_post_comment_in_usdb")
        icon11 = QIcon()
        icon11.addFile(u":/icons/balloon.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_post_comment_in_usdb.setIcon(icon11)
        self.action_open_song_folder = QAction(MainWindow)
        self.action_open_song_folder.setObjectName(u"action_open_song_folder")
        icon12 = QIcon()
        icon12.addFile(u":/icons/folder_note.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_folder.setIcon(icon12)
        self.action_delete = QAction(MainWindow)
        self.action_delete.setObjectName(u"action_delete")
        icon13 = QIcon()
        icon13.addFile(u":/icons/bin.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_delete.setIcon(icon13)
        self.action_pin = QAction(MainWindow)
        self.action_pin.setObjectName(u"action_pin")
        self.action_pin.setCheckable(True)
        icon14 = QIcon()
        icon14.addFile(u":/icons/pin.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_pin.setIcon(icon14)
        self.action_songs_abort = QAction(MainWindow)
        self.action_songs_abort.setObjectName(u"action_songs_abort")
        icon15 = QIcon()
        icon15.addFile(u":/icons/minus-circle.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_songs_abort.setIcon(icon15)
        self.action_delete_saved_search = QAction(MainWindow)
        self.action_delete_saved_search.setObjectName(u"action_delete_saved_search")
        self.action_delete_saved_search.setMenuRole(QAction.MenuRole.NoRole)
        self.action_add_saved_search = QAction(MainWindow)
        self.action_add_saved_search.setObjectName(u"action_add_saved_search")
        self.action_add_saved_search.setMenuRole(QAction.MenuRole.NoRole)
        self.action_update_saved_search = QAction(MainWindow)
        self.action_update_saved_search.setObjectName(u"action_update_saved_search")
        self.action_update_saved_search.setMenuRole(QAction.MenuRole.NoRole)
        self.action_set_saved_search_default = QAction(MainWindow)
        self.action_set_saved_search_default.setObjectName(u"action_set_saved_search_default")
        self.action_set_saved_search_default.setCheckable(True)
        self.action_set_saved_search_default.setMenuRole(QAction.MenuRole.NoRole)
        self.action_set_saved_search_subscribed = QAction(MainWindow)
        self.action_set_saved_search_subscribed.setObjectName(u"action_set_saved_search_subscribed")
        self.action_set_saved_search_subscribed.setCheckable(True)
        self.action_set_saved_search_subscribed.setMenuRole(QAction.MenuRole.NoRole)
        self.action_open_song_in_karedi = QAction(MainWindow)
        self.action_open_song_in_karedi.setObjectName(u"action_open_song_in_karedi")
        icon16 = QIcon()
        icon16.addFile(u":/icons/karedi.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_in_karedi.setIcon(icon16)
        self.action_open_song_in_performous = QAction(MainWindow)
        self.action_open_song_in_performous.setObjectName(u"action_open_song_in_performous")
        icon17 = QIcon()
        icon17.addFile(u":/icons/performous.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_in_performous.setIcon(icon17)
        self.action_open_song_in_usdx = QAction(MainWindow)
        self.action_open_song_in_usdx.setObjectName(u"action_open_song_in_usdx")
        icon18 = QIcon()
        icon18.addFile(u":/icons/usdx.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_in_usdx.setIcon(icon18)
        self.action_open_song_in_ultrastar_manager = QAction(MainWindow)
        self.action_open_song_in_ultrastar_manager.setObjectName(u"action_open_song_in_ultrastar_manager")
        icon19 = QIcon()
        icon19.addFile(u":/icons/ultrastar-manager.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_in_ultrastar_manager.setIcon(icon19)
        self.action_open_song_in_vocaluxe = QAction(MainWindow)
        self.action_open_song_in_vocaluxe.setObjectName(u"action_open_song_in_vocaluxe")
        icon20 = QIcon()
        icon20.addFile(u":/icons/vocaluxe.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_in_vocaluxe.setIcon(icon20)
        self.action_open_song_in_yass_reloaded = QAction(MainWindow)
        self.action_open_song_in_yass_reloaded.setObjectName(u"action_open_song_in_yass_reloaded")
        icon21 = QIcon()
        icon21.addFile(u":/icons/yass-reloaded.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_open_song_in_yass_reloaded.setIcon(icon21)
        self.action_rate_in_usdb = QAction(MainWindow)
        self.action_rate_in_usdb.setObjectName(u"action_rate_in_usdb")
        icon22 = QIcon()
        icon22.addFile(u":/icons/rating.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.action_rate_in_usdb.setIcon(icon22)
        self.action_rate_in_usdb.setMenuRole(QAction.MenuRole.NoRole)
        self.action_rate_1star = QAction(MainWindow)
        self.action_rate_1star.setObjectName(u"action_rate_1star")
        self.action_rate_2stars = QAction(MainWindow)
        self.action_rate_2stars.setObjectName(u"action_rate_2stars")
        self.action_rate_3stars = QAction(MainWindow)
        self.action_rate_3stars.setObjectName(u"action_rate_3stars")
        self.action_rate_4stars = QAction(MainWindow)
        self.action_rate_4stars.setObjectName(u"action_rate_4stars")
        self.action_rate_5stars = QAction(MainWindow)
        self.action_rate_5stars.setObjectName(u"action_rate_5stars")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout_3 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.lineEdit_search = QLineEdit(self.centralwidget)
        self.lineEdit_search.setObjectName(u"lineEdit_search")

        self.horizontalLayout_4.addWidget(self.lineEdit_search)

        self.label_update_hint = QLabel(self.centralwidget)
        self.label_update_hint.setObjectName(u"label_update_hint")

        self.horizontalLayout_4.addWidget(self.label_update_hint)


        self.verticalLayout_3.addLayout(self.horizontalLayout_4)

        self.table_view = QTableView(self.centralwidget)
        self.table_view.setObjectName(u"table_view")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(3)
        sizePolicy.setHeightForWidth(self.table_view.sizePolicy().hasHeightForWidth())
        self.table_view.setSizePolicy(sizePolicy)
        self.table_view.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.table_view.setAlternatingRowColors(True)
        self.table_view.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.table_view.setSortingEnabled(True)
        self.table_view.horizontalHeader().setMinimumSectionSize(20)
        self.table_view.horizontalHeader().setDefaultSectionSize(20)
        self.table_view.horizontalHeader().setHighlightSections(False)
        self.table_view.horizontalHeader().setStretchLastSection(True)
        self.table_view.verticalHeader().setVisible(False)
        self.table_view.verticalHeader().setMinimumSectionSize(24)
        self.table_view.verticalHeader().setDefaultSectionSize(24)

        self.verticalLayout_3.addWidget(self.table_view)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.button_download = QPushButton(self.centralwidget)
        self.button_download.setObjectName(u"button_download")
        self.button_download.setEnabled(True)
        self.button_download.setIcon(icon8)
        self.button_download.setIconSize(QSize(20, 20))

        self.horizontalLayout.addWidget(self.button_download)

        self.button_pause = QToolButton(self.centralwidget)
        self.button_pause.setObjectName(u"button_pause")
        icon23 = QIcon()
        icon23.addFile(u":/icons/control-pause.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.button_pause.setIcon(icon23)
        self.button_pause.setIconSize(QSize(20, 20))
        self.button_pause.setCheckable(True)

        self.horizontalLayout.addWidget(self.button_pause)

        self.bar_download_progress = QProgressBar(self.centralwidget)
        self.bar_download_progress.setObjectName(u"bar_download_progress")
        self.bar_download_progress.setValue(0)
        self.bar_download_progress.setTextVisible(False)

        self.horizontalLayout.addWidget(self.bar_download_progress)

        self.label_download_progress = QLabel(self.centralwidget)
        self.label_download_progress.setObjectName(u"label_download_progress")
        self.label_download_progress.setMinimumSize(QSize(100, 0))
        self.label_download_progress.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.horizontalLayout.addWidget(self.label_download_progress)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.pushButton_select_song_dir = QPushButton(self.centralwidget)
        self.pushButton_select_song_dir.setObjectName(u"pushButton_select_song_dir")
        self.pushButton_select_song_dir.setEnabled(True)
        self.pushButton_select_song_dir.setIcon(icon12)

        self.horizontalLayout.addWidget(self.pushButton_select_song_dir)

        self.lineEdit_song_dir = QLineEdit(self.centralwidget)
        self.lineEdit_song_dir.setObjectName(u"lineEdit_song_dir")
        self.lineEdit_song_dir.setReadOnly(True)

        self.horizontalLayout.addWidget(self.lineEdit_song_dir)


        self.verticalLayout_3.addLayout(self.horizontalLayout)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setEnabled(True)
        self.menubar.setGeometry(QRect(0, 0, 1540, 24))
        self.menu_tools = QMenu(self.menubar)
        self.menu_tools.setObjectName(u"menu_tools")
        self.menu_tools.setToolTipsVisible(True)
        self.menu_usdb = QMenu(self.menubar)
        self.menu_usdb.setObjectName(u"menu_usdb")
        self.menu_usdb.setToolTipsVisible(True)
        self.menu_about = QMenu(self.menubar)
        self.menu_about.setObjectName(u"menu_about")
        self.menu_about.setToolTipsVisible(True)
        self.menu_songs = QMenu(self.menubar)
        self.menu_songs.setObjectName(u"menu_songs")
        self.menu_songs.setToolTipsVisible(True)
        self.menu_custom_data = QMenu(self.menu_songs)
        self.menu_custom_data.setObjectName(u"menu_custom_data")
        icon24 = QIcon()
        icon24.addFile(u":/icons/drawer.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.menu_custom_data.setIcon(icon24)
        self.menu_open_song_in = QMenu(self.menu_songs)
        self.menu_open_song_in.setObjectName(u"menu_open_song_in")
        icon25 = QIcon()
        icon25.addFile(u":/icons/music--arrow.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.menu_open_song_in.setIcon(icon25)
        self.menu_rate_song_on_usdb = QMenu(self.menu_songs)
        self.menu_rate_song_on_usdb.setObjectName(u"menu_rate_song_on_usdb")
        self.menu_rate_song_on_usdb.setIcon(icon22)
        self.menu_local = QMenu(self.menubar)
        self.menu_local.setObjectName(u"menu_local")
        self.menu_local.setToolTipsVisible(True)
        self.menu_view = QMenu(self.menubar)
        self.menu_view.setObjectName(u"menu_view")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.dock_log = QDockWidget(MainWindow)
        self.dock_log.setObjectName(u"dock_log")
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName(u"dockWidgetContents")
        self.verticalLayout_2 = QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.toolButton_errors = QToolButton(self.dockWidgetContents)
        self.toolButton_errors.setObjectName(u"toolButton_errors")
        icon26 = QIcon()
        icon26.addFile(u":/icons/error.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_errors.setIcon(icon26)
        self.toolButton_errors.setIconSize(QSize(20, 20))
        self.toolButton_errors.setCheckable(True)
        self.toolButton_errors.setChecked(True)

        self.horizontalLayout_3.addWidget(self.toolButton_errors)

        self.toolButton_warnings = QToolButton(self.dockWidgetContents)
        self.toolButton_warnings.setObjectName(u"toolButton_warnings")
        icon27 = QIcon()
        icon27.addFile(u":/icons/warning.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_warnings.setIcon(icon27)
        self.toolButton_warnings.setIconSize(QSize(20, 20))
        self.toolButton_warnings.setCheckable(True)
        self.toolButton_warnings.setChecked(True)

        self.horizontalLayout_3.addWidget(self.toolButton_warnings)

        self.toolButton_infos = QToolButton(self.dockWidgetContents)
        self.toolButton_infos.setObjectName(u"toolButton_infos")
        icon28 = QIcon()
        icon28.addFile(u":/icons/info.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_infos.setIcon(icon28)
        self.toolButton_infos.setIconSize(QSize(20, 20))
        self.toolButton_infos.setCheckable(True)
        self.toolButton_infos.setChecked(True)

        self.horizontalLayout_3.addWidget(self.toolButton_infos)

        self.toolButton_debugs = QToolButton(self.dockWidgetContents)
        self.toolButton_debugs.setObjectName(u"toolButton_debugs")
        icon29 = QIcon()
        icon29.addFile(u":/icons/bug.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.toolButton_debugs.setIcon(icon29)
        self.toolButton_debugs.setIconSize(QSize(20, 20))
        self.toolButton_debugs.setCheckable(True)

        self.horizontalLayout_3.addWidget(self.toolButton_debugs)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer)


        self.verticalLayout_2.addLayout(self.horizontalLayout_3)

        self.plainTextEdit = QPlainTextEdit(self.dockWidgetContents)
        self.plainTextEdit.setObjectName(u"plainTextEdit")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.plainTextEdit.sizePolicy().hasHeightForWidth())
        self.plainTextEdit.setSizePolicy(sizePolicy1)

        self.verticalLayout_2.addWidget(self.plainTextEdit)

        self.dock_log.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(Qt.DockWidgetArea.RightDockWidgetArea, self.dock_log)
        self.dock_search = QDockWidget(MainWindow)
        self.dock_search.setObjectName(u"dock_search")
        self.dockWidgetContents_2 = QWidget()
        self.dockWidgetContents_2.setObjectName(u"dockWidgetContents_2")
        self.verticalLayout = QVBoxLayout(self.dockWidgetContents_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.line_edit_search_filters = QLineEdit(self.dockWidgetContents_2)
        self.line_edit_search_filters.setObjectName(u"line_edit_search_filters")
        self.line_edit_search_filters.setFrame(False)

        self.verticalLayout.addWidget(self.line_edit_search_filters)

        self.search_view = QTreeView(self.dockWidgetContents_2)
        self.search_view.setObjectName(u"search_view")
        self.search_view.setFrameShape(QFrame.Shape.NoFrame)

        self.verticalLayout.addWidget(self.search_view)

        self.dock_search.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self.dock_search)

        self.menubar.addAction(self.menu_view.menuAction())
        self.menubar.addAction(self.menu_songs.menuAction())
        self.menubar.addAction(self.menu_usdb.menuAction())
        self.menubar.addAction(self.menu_local.menuAction())
        self.menubar.addAction(self.menu_tools.menuAction())
        self.menubar.addAction(self.menu_about.menuAction())
        self.menu_tools.addAction(self.action_settings)
        self.menu_tools.addAction(self.action_meta_tags)
        self.menu_tools.addAction(self.action_generate_song_list)
        self.menu_usdb.addAction(self.action_usdb_login)
        self.menu_usdb.addAction(self.action_refetch_song_list)
        self.menu_about.addAction(self.action_show_log)
        self.menu_about.addAction(self.action_about)
        self.menu_songs.addAction(self.action_songs_download)
        self.menu_songs.addAction(self.action_songs_abort)
        self.menu_songs.addAction(self.action_show_in_usdb)
        self.menu_songs.addAction(self.action_post_comment_in_usdb)
        self.menu_songs.addAction(self.menu_rate_song_on_usdb.menuAction())
        self.menu_songs.addSeparator()
        self.menu_songs.addAction(self.action_open_song_folder)
        self.menu_songs.addAction(self.menu_open_song_in.menuAction())
        self.menu_songs.addAction(self.menu_custom_data.menuAction())
        self.menu_songs.addAction(self.action_pin)
        self.menu_songs.addSeparator()
        self.menu_songs.addAction(self.action_delete)
        self.menu_open_song_in.addAction(self.action_open_song_in_usdx)
        self.menu_open_song_in.addAction(self.action_open_song_in_vocaluxe)
        self.menu_open_song_in.addAction(self.action_open_song_in_performous)
        self.menu_open_song_in.addSeparator()
        self.menu_open_song_in.addAction(self.action_open_song_in_yass_reloaded)
        self.menu_open_song_in.addAction(self.action_open_song_in_karedi)
        self.menu_open_song_in.addAction(self.action_open_song_in_ultrastar_manager)
        self.menu_rate_song_on_usdb.addAction(self.action_rate_1star)
        self.menu_rate_song_on_usdb.addAction(self.action_rate_2stars)
        self.menu_rate_song_on_usdb.addAction(self.action_rate_3stars)
        self.menu_rate_song_on_usdb.addAction(self.action_rate_4stars)
        self.menu_rate_song_on_usdb.addAction(self.action_rate_5stars)
        self.menu_local.addAction(self.action_find_local_songs)
        self.menu_local.addAction(self.action_import_usdb_ids)
        self.menu_local.addAction(self.action_export_usdb_ids)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"USDB Syncer", None))
        self.action_meta_tags.setText(QCoreApplication.translate("MainWindow", u"Create Meta Tags", None))
        self.action_settings.setText(QCoreApplication.translate("MainWindow", u"Settings", None))
        self.action_generate_song_list.setText(QCoreApplication.translate("MainWindow", u"Create Report", None))
        self.action_refetch_song_list.setText(QCoreApplication.translate("MainWindow", u"Check USDB Song List", None))
#if QT_CONFIG(tooltip)
        self.action_refetch_song_list.setToolTip(QCoreApplication.translate("MainWindow", u"Redownload the entire song list from USDB", None))
#endif // QT_CONFIG(tooltip)
        self.action_find_local_songs.setText(QCoreApplication.translate("MainWindow", u"Find Local Songs", None))
#if QT_CONFIG(tooltip)
        self.action_find_local_songs.setToolTip(QCoreApplication.translate("MainWindow", u"Select songs matching a txt file in the given folder", None))
#endif // QT_CONFIG(tooltip)
        self.action_show_log.setText(QCoreApplication.translate("MainWindow", u"Show Log", None))
        self.action_import_usdb_ids.setText(QCoreApplication.translate("MainWindow", u"Select Songs from Files", None))
#if QT_CONFIG(tooltip)
        self.action_import_usdb_ids.setToolTip(QCoreApplication.translate("MainWindow", u"Read song IDs from selected files and select the corresponding songs in the song table", None))
#endif // QT_CONFIG(tooltip)
        self.action_export_usdb_ids.setText(QCoreApplication.translate("MainWindow", u"Export selected IDs to File", None))
#if QT_CONFIG(tooltip)
        self.action_export_usdb_ids.setToolTip(QCoreApplication.translate("MainWindow", u"Write the IDs of the selected songs to a file", None))
#endif // QT_CONFIG(tooltip)
        self.action_songs_download.setText(QCoreApplication.translate("MainWindow", u"Download", None))
        self.action_usdb_login.setText(QCoreApplication.translate("MainWindow", u"USDB Login", None))
        self.action_about.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.action_show_in_usdb.setText(QCoreApplication.translate("MainWindow", u"Show on USDB", None))
        self.action_post_comment_in_usdb.setText(QCoreApplication.translate("MainWindow", u"Post Comment on USDB", None))
        self.action_open_song_folder.setText(QCoreApplication.translate("MainWindow", u"Open Song Folder", None))
        self.action_delete.setText(QCoreApplication.translate("MainWindow", u"Delete", None))
#if QT_CONFIG(tooltip)
        self.action_delete.setToolTip(QCoreApplication.translate("MainWindow", u"Move to system recycle bin", None))
#endif // QT_CONFIG(tooltip)
        self.action_pin.setText(QCoreApplication.translate("MainWindow", u"Pin", None))
#if QT_CONFIG(tooltip)
        self.action_pin.setToolTip(QCoreApplication.translate("MainWindow", u"Disable or reenable updates to local files", None))
#endif // QT_CONFIG(tooltip)
        self.action_songs_abort.setText(QCoreApplication.translate("MainWindow", u"Abort Download", None))
        self.action_delete_saved_search.setText(QCoreApplication.translate("MainWindow", u"Delete", None))
        self.action_add_saved_search.setText(QCoreApplication.translate("MainWindow", u"Save current search", None))
        self.action_update_saved_search.setText(QCoreApplication.translate("MainWindow", u"Update with current search", None))
        self.action_set_saved_search_default.setText(QCoreApplication.translate("MainWindow", u"Apply on startup", None))
#if QT_CONFIG(tooltip)
        self.action_set_saved_search_default.setToolTip(QCoreApplication.translate("MainWindow", u"If checked, this search will automatically applied every time you start the application.", None))
#endif // QT_CONFIG(tooltip)
        self.action_set_saved_search_subscribed.setText(QCoreApplication.translate("MainWindow", u"Download new matches", None))
        self.action_open_song_in_karedi.setText(QCoreApplication.translate("MainWindow", u"Karedi", None))
        self.action_open_song_in_performous.setText(QCoreApplication.translate("MainWindow", u"Performous", None))
        self.action_open_song_in_usdx.setText(QCoreApplication.translate("MainWindow", u"UltraStar Deluxe", None))
        self.action_open_song_in_ultrastar_manager.setText(QCoreApplication.translate("MainWindow", u"UltraStar Manager", None))
        self.action_open_song_in_vocaluxe.setText(QCoreApplication.translate("MainWindow", u"Vocaluxe", None))
        self.action_open_song_in_yass_reloaded.setText(QCoreApplication.translate("MainWindow", u"YASS Reloaded", None))
        self.action_rate_in_usdb.setText(QCoreApplication.translate("MainWindow", u"action_rate_in_usdb", None))
        self.action_rate_1star.setText(QCoreApplication.translate("MainWindow", u"\u2605 \u2013 Unusable, requires a complete overhaul", None))
#if QT_CONFIG(tooltip)
        self.action_rate_1star.setToolTip(QCoreApplication.translate("MainWindow", u"unusable, needs a total rework", None))
#endif // QT_CONFIG(tooltip)
        self.action_rate_2stars.setText(QCoreApplication.translate("MainWindow", u"\u2605\u2605 \u2013 Barely usable, needs significant revisions", None))
#if QT_CONFIG(tooltip)
        self.action_rate_2stars.setToolTip(QCoreApplication.translate("MainWindow", u"hardly usable, needs a lot of improvements", None))
#endif // QT_CONFIG(tooltip)
        self.action_rate_3stars.setText(QCoreApplication.translate("MainWindow", u"\u2605\u2605\u2605 \u2013 Usable, but requires improvements", None))
#if QT_CONFIG(tooltip)
        self.action_rate_3stars.setToolTip(QCoreApplication.translate("MainWindow", u"usable, but room for improvement", None))
#endif // QT_CONFIG(tooltip)
        self.action_rate_4stars.setText(QCoreApplication.translate("MainWindow", u"\u2605\u2605\u2605\u2605 \u2013 Very good, with minor areas for improvement", None))
#if QT_CONFIG(tooltip)
        self.action_rate_4stars.setToolTip(QCoreApplication.translate("MainWindow", u"room for some improvements", None))
#endif // QT_CONFIG(tooltip)
        self.action_rate_5stars.setText(QCoreApplication.translate("MainWindow", u"\u2605\u2605\u2605\u2605\u2605 \u2013 Excellent, no improvements needed", None))
#if QT_CONFIG(tooltip)
        self.action_rate_5stars.setToolTip(QCoreApplication.translate("MainWindow", u"simply perfect", None))
#endif // QT_CONFIG(tooltip)
        self.lineEdit_search.setPlaceholderText(QCoreApplication.translate("MainWindow", u"Search for anything ...", None))
        self.label_update_hint.setText(QCoreApplication.translate("MainWindow", u"<html><head/><body><p><a href=\"https://github.com/bohning/usdb_syncer/releases\"><span style=\" text-decoration: underline; color:#3586ff;\">USDB Syncer VERSION available!</span></a></p></body></html>", None))
        self.button_download.setText(QCoreApplication.translate("MainWindow", u"Download", None))
        self.button_pause.setText(QCoreApplication.translate("MainWindow", u"Pause", None))
        self.label_download_progress.setText(QCoreApplication.translate("MainWindow", u"0/0", None))
        self.pushButton_select_song_dir.setText(QCoreApplication.translate("MainWindow", u"Select song folder", None))
        self.lineEdit_song_dir.setText(QCoreApplication.translate("MainWindow", u"./songs", None))
        self.menu_tools.setTitle(QCoreApplication.translate("MainWindow", u"Tools", None))
        self.menu_usdb.setTitle(QCoreApplication.translate("MainWindow", u"USDB", None))
        self.menu_about.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))
        self.menu_songs.setTitle(QCoreApplication.translate("MainWindow", u"Songs", None))
        self.menu_custom_data.setTitle(QCoreApplication.translate("MainWindow", u"Custom Data", None))
        self.menu_open_song_in.setTitle(QCoreApplication.translate("MainWindow", u"Open Song in", None))
        self.menu_rate_song_on_usdb.setTitle(QCoreApplication.translate("MainWindow", u"Rate Song on USDB", None))
        self.menu_local.setTitle(QCoreApplication.translate("MainWindow", u"Local", None))
        self.menu_view.setTitle(QCoreApplication.translate("MainWindow", u"View", None))
        self.dock_log.setWindowTitle(QCoreApplication.translate("MainWindow", u"Log", None))
#if QT_CONFIG(tooltip)
        self.toolButton_errors.setToolTip(QCoreApplication.translate("MainWindow", u"Errors", None))
#endif // QT_CONFIG(tooltip)
        self.toolButton_errors.setText(QCoreApplication.translate("MainWindow", u"Errors", None))
#if QT_CONFIG(tooltip)
        self.toolButton_warnings.setToolTip(QCoreApplication.translate("MainWindow", u"Warnings", None))
#endif // QT_CONFIG(tooltip)
        self.toolButton_warnings.setText(QCoreApplication.translate("MainWindow", u"Warnings", None))
#if QT_CONFIG(tooltip)
        self.toolButton_infos.setToolTip(QCoreApplication.translate("MainWindow", u"Information", None))
#endif // QT_CONFIG(tooltip)
        self.toolButton_infos.setText(QCoreApplication.translate("MainWindow", u"Infos", None))
        self.toolButton_debugs.setText(QCoreApplication.translate("MainWindow", u"Debugs", None))
        self.dock_search.setWindowTitle(QCoreApplication.translate("MainWindow", u"Filters", None))
        self.line_edit_search_filters.setPlaceholderText(QCoreApplication.translate("MainWindow", u"Search filters", None))
    # retranslateUi

